#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 22:00:00                  #
# ================================================== #

from typing import Dict, Any, Optional

from packaging.version import Version

from pygpt_net.item.index import IndexItem
from pygpt_net.provider.core.index.base import BaseProvider
from .patch import Patch
from .storage import Storage


class DbSqliteProvider(BaseProvider):
    def __init__(self, window=None):
        super(DbSqliteProvider, self).__init__(window)
        self.window = window
        self.patcher = Patch(window, self)
        self.storage = Storage(window)
        self.id = "db_sqlite"
        self.type = "index"

    def attach(self, window):
        """
        Attach window to provider

        :param window: Window instance
        """
        self.window = window
        self.storage.attach(window)

    def patch(self, version: Version):
        """
        Patch versions

        :param version: current app version
        """
        return self.patcher.execute(version)

    def load(
            self,
            store_id: str
    ) -> Dict[str, IndexItem]:
        """
        Load items for index provider

        :param store_id: store ID
        :return: dict of indexes and files in index
        """
        return self.storage.get_items(store_id)

    def append_file(
            self,
            store_id: str,
            idx: str,
            data: Dict[str, Any]
    ) -> int:
        """
        Append file to index

        :param: store_id: store ID
        :param: idx: index
        :param: data: dict of file data
        :return: ID of inserted file
        """
        return self.storage.insert_file(store_id, idx, data)

    def append_ctx_meta(
            self,
            store_id: str,
            idx: str,
            meta_id: int,
            doc_id: str
    ) -> int:
        """
        Append context meta to index

        :param: store_id: store ID
        :param: idx: index
        :param: meta_id: context meta ID
        :param: doc_id: document ID
        :return: ID of inserted context meta
        """
        return self.storage.insert_ctx_meta(store_id, idx, meta_id, doc_id)

    def append_external(
            self,
            store_id: str,
            idx: str,
            data: Dict[str, Any]
    ) -> int:
        """
        Append external data to index

        :param: store_id: store ID
        :param: idx: index
        :param: data: dict of external data
        :return: ID of inserted external
        """
        return self.storage.insert_external(store_id, idx, data)

    def is_meta_indexed(
            self,
            store_id: str,
            idx: str,
            meta_id: int
    ) -> bool:
        """
        Check if context meta is indexed

        :param: store_id: store ID
        :param: idx: index
        :param: meta_id: context meta ID
        :return: True if indexed
        """
        return self.storage.is_meta_indexed(store_id, idx, meta_id)

    def is_file_indexed(
            self,
            store_id: str,
            idx: str,
            file_id: str
    ) -> bool:
        """
        Check if file is indexed

        :param: store_id: store ID
        :param: idx: index
        :param: file_id: file ID
        :return: True if indexed
        """
        return self.storage.is_file_indexed(store_id, idx, file_id)

    def is_external_indexed(
            self,
            store_id: str,
            idx: str,
            content: str,
            type: str
    ) -> bool:
        """
        Check if external data is indexed

        :param: store_id: store ID
        :param: idx: index
        :param: content: external content
        :param: type: external type
        :return: True if indexed
        """
        return self.storage.is_external_indexed(store_id, idx, content, type)

    def get_meta_doc_id(
            self,
            store_id: str,
            idx: str,
            meta_id: int
    ) -> str:
        """
        Get indexed document id by meta id

        :param store_id: store id
        :param idx: index name
        :param meta_id: meta id
        :return: document id
        """
        return self.storage.get_meta_doc_id(store_id, idx, meta_id)

    def get_file_doc_id(
            self,
            store_id: str,
            idx: str,
            file_id: str
    ) -> str:
        """
        Get indexed document id by file id

        :param store_id: store id
        :param idx: index name
        :param file_id: file id
        :return: document id
        """
        return self.storage.get_file_doc_id(store_id, idx, file_id)

    def get_external_doc_id(
            self,
            store_id: str,
            idx: str,
            content: str,
            type: str
    ) -> str:
        """
        Get indexed document id by file id

        :param store_id: store id
        :param idx: index name
        :param content: external content
        :param type: external type
        :return: document id
        """
        return self.storage.get_external_doc_id(store_id, idx, content, type)

    def update_file(
            self,
            id: int,
            doc_id: str,
            ts: int
    ) -> bool:
        """
        Update indexed timestamp of indexed file

        :param: id: db record ID
        :param: doc_id: document ID
        :param: ts: timestamp
        """
        return self.storage.update_file(id, doc_id, ts)

    def update_ctx_meta(
            self,
            meta_id: int,
            doc_id: str
    ) -> bool:
        """
        Update indexed timestamp of indexed context meta

        :param: meta_id: context meta ID
        :param: doc_id: document ID
        """
        return self.storage.update_ctx_meta(meta_id, doc_id)

    def update_external(
            self,
            content: str,
            type: str,
            doc_id: str,
            ts: int
    ) -> bool:
        """
        Update indexed timestamp of indexed external

        :param: content: external content
        :param: type: external type
        :param: doc_id: document ID
        :param: ts: timestamp
        """
        return self.storage.update_external(content, type, doc_id, ts)

    def remove_file(
            self,
            store_id: str,
            idx: str,
            doc_id: str
    ):
        """
        Remove file from index

        :param store_id: store ID
        :param idx: index
        :param doc_id: document ID
        """
        self.storage.remove_file(store_id, idx, doc_id)

    def remove_ctx_meta(
            self,
            store_id: str,
            idx: str,
            meta_id: str
    ):
        """
        Remove file from index

        :param store_id: store ID
        :param idx: index
        :param meta_id: context meta ID
        """
        self.storage.remove_ctx_meta(store_id, idx, meta_id)

    def remove_external(
            self,
            store_id: str,
            idx: str,
            doc_id: str
    ):
        """
        Remove external from index

        :param store_id: store ID
        :param idx: index
        :param doc_id: document ID
        """
        self.storage.remove_external(store_id, idx, doc_id)

    def truncate(
            self,
            store_id: str,
            idx: str
    ) -> bool:
        """
        Truncate idx (remove all items)

        :param store_id: store ID
        :param idx: index
        :return: True if truncated
        """
        return self.storage.truncate_all(store_id, idx)

    def truncate_all(
            self,
            store_id: Optional[str] = None,
            idx: Optional[str] = None
    ) -> bool:
        """
        Truncate idx (remove all items) - all store and indexes

        :param store_id: store ID
        :param idx: index
        :return: True if truncated
        """
        return self.storage.truncate_all(store_id, idx)

    def truncate_files(
            self,
            store_id: Optional[str] = None,
            idx: Optional[str] = None
    ) -> bool:
        """
        Truncate files idx (remove all items) - all store and indexes

        :param store_id: store ID
        :param idx: index
        :return: True if truncated
        """
        return self.storage.truncate_files(store_id, idx)

    def truncate_ctx(
            self,
            store_id: Optional[str] = None,
            idx: Optional[str] = None
    ) -> bool:
        """
        Truncate context meta idx (remove all items) - all store and indexes

        :param store_id: store ID
        :param idx: index
        :return: True if truncated
        """
        return self.storage.truncate_ctx(store_id, idx)

    def truncate_external(
            self,
            store_id: Optional[str] = None,
            idx: Optional[str] = None
    ) -> bool:
        """
        Truncate external idx (remove all items) - all store and indexes

        :param store_id: store ID
        :param idx: index
        :return: True if truncated
        """
        return self.storage.truncate_external(store_id, idx)

    def get_counters(
            self,
            type: str
    ) -> Dict[str, Dict[str, int]]:
        """
        Get counters (stats, count items by type [file, ctx, external])

        :param type: type of counter (file, ctx, external)
        :return: dict of counters
        """
        return self.storage.get_counters(type)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.02.02 02:00:00                  #
# ================================================== #

import json
import os
import shutil
from typing import Optional, Dict, Any

from packaging.version import Version

from pygpt_net.provider.core.model.base import BaseProvider
from pygpt_net.item.model import ModelItem

from .patch import Patch


class JsonFileProvider(BaseProvider):
    def __init__(self, window=None):
        super(JsonFileProvider, self).__init__(window)
        self.window = window
        self.patcher = Patch(window)
        self.id = "json_file"
        self.type = "model"
        self.config_file = 'models.json'

    def install(self):
        """Install provider data files"""
        dst = os.path.join(self.window.core.config.path, self.config_file)
        if not os.path.exists(dst):
            src = os.path.join(self.window.core.config.get_app_path(), 'data', 'config', self.config_file)
            shutil.copyfile(src, dst)

    def get_version(self) -> Optional[str]:
        """
        Get data version

        :return: version
        """
        path = os.path.join(self.window.core.config.path, self.config_file)
        with open(path, 'r', encoding="utf-8") as file:
            data = json.load(file)
            if data == "" or data is None:
                return
            if '__meta__' in data and 'version' in data['__meta__']:
                return data['__meta__']['version']

    def load_base(self) -> Optional[Dict[str, ModelItem]]:
        """
        Load base models config from base JSON file

        :return: models dict
        """
        path = os.path.join(self.window.core.config.get_app_path(), 'data', 'config', self.config_file)
        return self.load(path)

    def load(self, path: Optional[str] = None) -> Optional[Dict[str, ModelItem]]:
        """
        Load models config from JSON file
        """
        items = {}
        if path is None:
            path = os.path.join(self.window.core.config.path, self.config_file)

        if not os.path.exists(path):
            print("FATAL ERROR: {} not found!".format(path))
            return None
        try:
            with open(path, 'r', encoding="utf-8") as file:
                data = json.load(file)
                if data == "" or data is None:
                    return {}

                # migrate from old versions < 2.0.49
                if 'items' not in data:
                    for id in data:
                        if id == '__meta__':
                            continue
                        item = data[id]
                        model = ModelItem()
                        self.deserialize(item, model)
                        items[id] = model
                    items = dict(sorted(items.items(), key=lambda item: item[0]))  # sort by key
                    print("Loaded models: {}".format(path))
                    print("Migrating old version: {}".format(path))
                    self.save(items)
                    return items

                # deserialize
                for id in data['items']:
                    item = data['items'][id]
                    model = ModelItem()
                    self.deserialize(item, model)
                    items[id] = model
                print("Loaded models: {}".format(path))

        except Exception as e:
            self.window.core.debug.log(e)

        return items

    def save(self, items: Dict[str, ModelItem]):
        """
        Save models config to JSON file

        :param items: models dict
        """
        path = os.path.join(self.window.core.config.path, self.config_file)
        try:
            data = {}
            ary = {}

            # serialize
            for id in items:
                model = items[id]
                ary[id] = self.serialize(model)

            data['__meta__'] = self.window.core.config.append_meta()
            data['items'] = ary
            dump = json.dumps(data, indent=4)
            with open(path, 'w', encoding="utf-8") as f:
                f.write(dump)

        except Exception as e:
            self.window.core.debug.log(e)

    def remove(self, id: str):
        pass

    def truncate(self):
        pass

    def patch(self, version: Version) -> bool:
        """
        Migrate models to current app version

        :param version: current app version
        :return: True if updated
        """
        return self.patcher.execute(version)

    @staticmethod
    def serialize(item: ModelItem) -> Dict[str, Any]:
        """
        Serialize item to dict

        :param item: item to serialize
        :return: serialized item
        """
        return {
            'id': item.id,
            'name': item.name,
            'mode': item.mode,
            'langchain': item.langchain,
            'llama_index': item.llama_index,
            'ctx': item.ctx,
            'tokens': item.tokens,
            'default': item.default,
            'multimodal': item.multimodal,
            'extra': item.extra,
        }

    @staticmethod
    def deserialize(data: Dict[str, Any], item: ModelItem):
        """
        Deserialize item from dict

        :param data: serialized item
        :param item: item to deserialize
        """
        if 'id' in data:
            item.id = data['id']
        if 'name' in data:
            item.name = data['name']
        if 'mode' in data:
            item.mode = data['mode']
        if 'langchain' in data:
            item.langchain = data['langchain']
        if 'llama_index' in data:
            item.llama_index = data['llama_index']
        if 'ctx' in data:
            item.ctx = data['ctx']
        if 'tokens' in data:
            item.tokens = data['tokens']
        if 'default' in data:
            item.default = data['default']
        if 'multimodal' in data:
            item.multimodal = data['multimodal']
        if 'extra' in data:
            item.extra = data['extra']

    def dump(self, item: ModelItem) -> str:
        """
        Dump to string

        :param item: item to dump
        :return: dumped item as string (json)
        """
        return json.dumps(self.serialize(item))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.03.02 19:00:00                  #
# ================================================== #

from typing import Dict, List

from pygpt_net.core.types import (
    MODE_AGENT,
    MODE_CHAT,
    MODE_COMPLETION,
    MODE_EXPERT,
    MODE_LANGCHAIN,
    MODE_LLAMA_INDEX,
    MODE_VISION,
    MODE_AUDIO,
    MODE_RESEARCH,
)
from pygpt_net.core.bridge.context import BridgeContext
from pygpt_net.core.events import Event, KernelEvent, RenderEvent
from pygpt_net.item.ctx import CtxItem
from pygpt_net.item.preset import PresetItem


class Experts:
    def __init__(self, window=None):
        """
        Experts core

        :param window: Window instance
        """
        self.window = window
        self.allowed_modes = [
            MODE_CHAT,
            MODE_COMPLETION,
            MODE_VISION,
            MODE_LANGCHAIN,
            MODE_LLAMA_INDEX,
            MODE_AUDIO,
            MODE_RESEARCH,
        ]
        self.allowed_cmds = ["expert_call"]

    def get_mode(self) -> str:
        """
        Get sub-mode to use internally

        :return: sub-mode
        """
        mode = MODE_CHAT
        current = self.window.core.config.get("experts.mode")
        if current is not None and current in self.allowed_modes:
            mode = current
        return mode

    def stopped(self) -> bool:
        """
        Check if experts are stopped

        :return: True if stopped
        """
        return self.window.controller.kernel.stopped()

    def agent_enabled(self) -> bool:
        """
        Check if agent is enabled

        :return: True if enabled
        """
        return self.window.controller.agent.legacy.enabled()

    def exists(self, id: str) -> bool:
        """
        Check if expert exists

        :param id: expert id
        :return: True if exists
        """
        return self.window.core.presets.has(MODE_EXPERT, id)

    def get_expert(self, id: str) -> PresetItem:
        """
        Get expert by id

        :param id: expert id
        :return: expert item (preset)
        """
        return self.window.core.presets.get_by_id(MODE_EXPERT, id)

    def get_experts(self) -> Dict[str, str]:
        """
        Get experts names with keys

        :return: experts dict
        """
        experts = {}
        presets = self.window.core.presets.get_by_mode(MODE_EXPERT)

        # mode: agent
        if self.agent_enabled():
            agents = self.window.core.presets.get_by_mode(MODE_AGENT)
            agent = self.window.core.config.get('preset')
            if agent is not None:
                if agent in agents:
                    for uuid in agents[agent].experts:
                        expert = self.window.core.presets.get_by_uuid(uuid)
                        if expert is not None:
                            id = expert.filename
                            experts[id] = expert.name
        # mode: expert
        else:
            for k in presets:
                if k.startswith("current."):  # skip current
                    continue
                if not presets[k].enabled:  # skip disabled experts
                    continue
                experts[k] = presets[k].name
        return experts

    def count_experts(self, uuid: str) -> int:
        """
        Count experts in agent

        :param uuid: agent uuid
        :return: number of experts
        """
        i = 0
        agents = self.window.core.presets.get_by_mode(MODE_AGENT)
        if uuid in agents:
            for expert_uuid in agents[uuid].experts:
                expert = self.window.core.presets.get_by_uuid(expert_uuid)
                if expert is not None:
                    i += 1
        return i

    def get_prompt(self) -> str:
        """
        Get prompt for handle experts

        :return: prompt with list of experts appended
        """
        prompt = self.window.core.config.get('prompt.expert')
        if self.window.core.command.is_native_enabled():
            prompt = """
            EXPERT MODE IS ENABLED
            ======================
            You are the professional manager of a team of experts, each of whom is the best in their respective fields. 
            
            IMPORTANT RULES:
            
            YOU MUST ALWAYS FOLLOW THIS RULES WHILE MANAGING EXPERTS:
            
            1. Instead of directly answering questions, you can manage these experts by forwarding queries to them, 
            collecting their responses, and controlling the flow of information between them as well as sending 
            requests to them. 
            2. You will serve as a bridge between them. 
            3. Each expert has their own conversation context, so you can communicate with them normally. 
            4. The experts can communicate with you but not with each other, so you will be responsible for relaying 
            information between them. 
            5. After consulting with the experts, your task will be to prepare the best possible answer based on the 
            assistance received and to use it in your Chain-of-Thought. 
            6. To send a request to an expert prepare the "expert_call" function call.
            7. Remember that the call must contain a query and the ID of the expert to be forwarded to the expert. 
            8. Use ONLY experts listed below, DO NOT TRY to ask any non-existing experts - they will not respond.
            9. Do not respond to the expert's replies, or use commands again if replying; the expert only has 
            the ability to listen to your queries sent and to respond to them.
            10. Always follow ALL user commands, even if you thing than expert can't do something. 
            Experts are allowed to execute commands same as you can.
            11. After query an expert, always wait for expert response before sending another query.
            12. Do not try to query experts in the middle of the conversation, put your queries and commands at the end of the text.
            13. Always call only one expert at a time.
            14. Always execute only one function call at a time.
            15. Always use the language used by user in conversation.
            16. Never write as an called expert and do not imagine response from him.
            17. Your response cannot include ~###~{...}~###~ commands in the conversation body - this syntax is reserved only for response purposes.
            18. First message after your query is always a response from asked expert, not from me.
            
            List of available experts, along with their IDs:
            
            {presets}
            """
        experts = self.get_experts()
        experts_list = []
        for k in experts:
            if k.startswith("current."): # skip current presets
                continue
            experts_list.append(" - " + str(k) + ": " + str(experts[k]))
        prompt = prompt.replace("{presets}", "\n".join(experts_list))
        return prompt

    def extract_calls(self, ctx: CtxItem) -> Dict[str, str]:
        """
        Extract expert calls from context output

        :param ctx: context item
        :return: dict with calls
        """
        ids = self.get_experts().keys()
        if not ids:  # abort if no experts
            return {}
        cmds = self.window.core.command.extract_cmds(ctx.output)
        if len(cmds) > 0:
            ctx.cmds = cmds  # append commands to ctx
        else:  # abort if no cmds
            return {}
        commands = self.window.core.command.from_commands(cmds)  # pack to execution list
        is_cmd = False
        my_commands = []
        calls = {}
        for item in commands:
            if item["cmd"] in self.allowed_cmds:
                my_commands.append(item)
                is_cmd = True
        if not is_cmd:  # abort if no expert calls
            return {}
        for item in my_commands:
            try:
                if item["cmd"] == "expert_call":
                    if "params" not in item:
                        continue
                    if "id" not in item["params"] or "query" not in item["params"]:
                        continue
                    id = item["params"]["id"]
                    if id not in ids:
                        continue
                    query = item["params"]["query"]
                    calls[id] = query
            except Exception as e:
                self.window.core.debug.error(e)
                return {}
        return calls

    def reply(self, ctx: CtxItem):
        """
        Re-send response from commands to master expert

        If command has been called by expert then response for command is send here to parent

        :param ctx: context item
        """
        if self.stopped():
            return

        # master meta is here, ctx.meta = MASTER META ID here!

        # make copy of ctx for reply, and change input name to expert name
        reply_ctx = CtxItem()
        reply_ctx.from_dict(ctx.to_dict())

        # reply ctx has no meta here!!!!
        reply_ctx.input_name = "Expert"
        reply_ctx.output_name = ""
        reply_ctx.sub_call = True  # this flag is not copied in to_dict

        internal = False
        if self.agent_enabled():  # hide in agent mode
            internal = True
        if ctx.output.strip() != "":
            response = reply_ctx.output
        else:
            response = reply_ctx.input

        context = BridgeContext()
        context.ctx = reply_ctx
        context.prompt = "Result from expert:\n\n" + str(response)
        extra = {
            "force": True,
            "reply": True,
            "internal": internal,
        }
        event = KernelEvent(KernelEvent.INPUT_SYSTEM, {
            'context': context,
            'extra': extra,
        })
        self.window.dispatch(event)

    def call(
            self,
            master_ctx: CtxItem,
            expert_id: str,
            query: str
    ):
        """
        Call the expert

        :param master_ctx: master context
        :param expert_id: expert id (preset ID)
        :param query: input text (master prompt)
        """
        if self.stopped():
            return

        # get or create children (slave) meta
        slave = self.window.core.ctx.get_or_create_slave_meta(master_ctx, expert_id)
        expert = self.get_expert(expert_id)
        reply = True
        hidden = False
        internal = False

        if self.agent_enabled():  # hide in agent mode
            internal = False
            hidden = True

        mode = self.window.core.config.get("mode")
        base_mode = mode
        model = expert.model
        expert_name = expert.name
        ai_name = ""
        sys_prompt = expert.prompt
        model_data = self.window.core.models.get(model)

        files = []
        file_ids = []
        functions = []
        tools_outputs = []

        # from current config
        max_tokens = self.window.core.config.get('max_output_tokens')
        stream_mode = self.window.core.config.get('stream')
        db_idx = self.window.controller.idx.current_idx # get idx from agent config

        mode = MODE_EXPERT  # force expert mode, mode will change in bridge

        # create slave item
        ctx = CtxItem()
        ctx.meta = slave  # use slave-meta
        ctx.internal = internal
        ctx.hidden = hidden
        ctx.current = True  # mark as current context item
        ctx.mode = mode  # store current selected mode (not inline changed)
        ctx.model = model  # store model list key, not real model id
        ctx.set_input(query, str(ai_name))
        ctx.set_output(None, expert_name)
        ctx.sub_call = True  # mark as sub-call
        ctx.pid = master_ctx.pid  # copy PID from parent to allow reply

        # render: begin
        event = RenderEvent(RenderEvent.BEGIN, {
            "meta": ctx.meta,
            "ctx": ctx,
            "stream": stream_mode,
        })
        self.window.dispatch(event)
        self.window.core.ctx.provider.append_item(slave, ctx)  # to slave meta

        # build sys prompt
        sys_prompt_raw = sys_prompt  # store raw prompt
        event = Event(Event.PRE_PROMPT, {
            'mode': mode,
            'value': sys_prompt,
            'is_expert': True,
        })
        self.window.dispatch(event)
        sys_prompt = event.data['value']
        sys_prompt = self.window.core.prompt.prepare_sys_prompt(
            mode,
            model_data,
            sys_prompt,
            ctx,
            reply,
            internal,
            is_expert=True,  # mark as expert, blocks expert prompt append in plugin
        )

        # call bridge
        history = self.window.core.ctx.all(
            meta_id=slave.id
        )  # get history for slave ctx
        bridge_context = BridgeContext(
            ctx=ctx,
            history=history,
            mode=mode,
            parent_mode=base_mode,
            model=model_data,
            system_prompt=sys_prompt,
            system_prompt_raw=sys_prompt_raw,
            prompt=query,
            stream=False,
            attachments=files,
            file_ids=file_ids,
            assistant_id=self.window.core.config.get('assistant'),
            idx=db_idx,
            idx_mode=self.window.core.config.get('llama.idx.mode'),
            external_functions=functions,
            tools_outputs=tools_outputs,
            max_tokens=max_tokens,
        )
        self.window.controller.chat.common.lock_input()  # lock input
        event = KernelEvent(KernelEvent.CALL, {
            'context': bridge_context,  # call using slave ctx history
            'extra': {},
        })
        self.window.dispatch(event)
        result = event.data.get("response")

        if not result:  # abort if bridge call failed
            return

        # handle output
        ctx.current = False  # reset current state
        ctx.output = result  # store expert output in their context
        
        self.window.core.ctx.update_item(ctx)

        ctx.from_previous()  # append previous result if exists

        # tmp switch meta for render purposes
        ctx.meta = master_ctx.meta
        self.window.controller.chat.output.handle(
            ctx=ctx,
            mode=mode,
            stream_mode=False,
        )
        ctx.clear_reply()  # reset results
        ctx.meta = slave  # restore before cmd execute

        self.window.controller.chat.command.handle(ctx)  # handle cmds
        self.window.controller.kernel.stack.handle()  # handle command queue

        # if command to execute then end here, and reply is returned to reply() above from stack, and ctx.reply = TRUE here
        # 
        ctx.from_previous()  # append previous result again before save
        self.window.core.ctx.update_item(ctx)  # update ctx in DB

        # if commands reply after bridge call, then stop (already handled in sync dispatcher)
        if ctx.reply:
            return

        # make copy of ctx for reply, and change input name to expert name
        reply_ctx = CtxItem()
        
        reply_ctx.from_dict(ctx.to_dict())
        reply_ctx.meta = master_ctx.meta

        # assign expert output
        reply_ctx.output = result
        reply_ctx.input_name = expert_name
        reply_ctx.output_name = ""
        reply_ctx.cmds = []  # clear cmds
        reply_ctx.sub_call = True  # this flag is not copied in to_dict

        # only if no command call, return final result to main
        context = BridgeContext()
        context.ctx = reply_ctx
        context.prompt = "@"+expert_id+" says:\n\n" + str(reply_ctx.output)
        extra = {
            "force": True,
            "reply": False,
            "internal": False,
        }
        event = KernelEvent(KernelEvent.INPUT_SYSTEM, {
            'context': context,
            'extra': extra,
        })
        self.window.dispatch(event)

    def get_functions(self) -> List[Dict[str, str]]:
        """
        Append call the expert commands

        :return: call the expert commands
        """
        cmds = [
            {
                "cmd": "expert_call",
                "instruction": "Call an expert",
                "params": [
                    {
                        "name": "id",
                        "description": "expert id",
                        "required": True,
                        "type": "str",
                    },
                    {
                        "name": "query",
                        "description": "query to expert",
                        "required": True,
                        "type": "str",
                    }
                ]
            }
        ]
        return cmds

    def has_calls(self, ctx: CtxItem) -> bool:
        """
        Check if context has expert calls

        :param ctx: CtxItem
        """
        if not ctx.sub_reply and not ctx.reply:
            mentions = self.window.core.experts.extract_calls(ctx)
            if mentions:
                for expert_id in mentions:
                    if not self.window.core.experts.exists(expert_id):
                        continue
                    return True
        return False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 17:00:00                  #
# ================================================== #

class TabsDebug:
    def __init__(self, window=None):
        """
        Tabs debug

        :param window: Window instance
        """
        self.window = window
        self.id = 'tabs'

    def update(self):
        """Update debug window."""
        self.window.core.debug.begin(self.id)
        self.window.core.debug.add(self.id, 'current Col', str(self.window.controller.ui.tabs.get_current_column_idx()))
        self.window.core.debug.add(self.id, 'current IDX', str(self.window.controller.ui.tabs.get_current_idx()))
        self.window.core.debug.add(self.id, 'current Tab', str(self.window.controller.ui.tabs.get_current_tab()))
        self.window.core.debug.add(self.id, 'current PID', str(self.window.controller.ui.tabs.get_current_pid()))
        self.window.core.debug.add(self.id, 'current Type', str(self.window.controller.ui.tabs.get_current_type()))
        self.window.core.debug.add(self.id, '----', '')
        self.window.core.debug.add(self.id, 'last_pid', str(self.window.core.tabs.last_pid))
        self.window.core.debug.add(self.id, 'locked', str(self.window.controller.ui.tabs.locked))
        self.window.core.debug.add(self.id, 'col', str(self.window.controller.ui.tabs.col))
        self.window.core.debug.add(self.id, 'count(pids)', str(len(self.window.core.tabs.pids)))
        self.window.core.debug.add(self.id, 'count(ctx bags)', str(len(self.window.core.ctx.container.bags)))
        self.window.core.debug.add(self.id, '----', '')

        # PIDs
        for pid in list(self.window.core.tabs.pids):
            tab = self.window.core.tabs.pids[pid]
            data = tab.to_dict()
            self.window.core.debug.add(self.id, "PID ["+str(pid)+"]", str(data))

        # mapping PID => meta.id
        self.window.core.debug.add(self.id, '----', '')
        self.window.core.debug.add(self.id, 'PID => meta.id', str(self.window.core.ctx.output.mapping))
        self.window.core.debug.add(self.id, '(last) meta.id => PID', str(self.window.core.ctx.output.last_pids))
        self.window.core.debug.add(self.id, '(last) PID', str(self.window.core.ctx.output.last_pid))
        self.window.core.debug.end(self.id)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.19 03:00:00                  #
# ================================================== #

from PySide6.QtCore import Slot, Signal

from pygpt_net.plugin.base.worker import BaseWorker, BaseSignals


class WorkerSignals(BaseSignals):
    output = Signal(object, str)
    html_output = Signal(object)
    ipython_output = Signal(object)
    build_finished = Signal()
    clear = Signal()


class Worker(BaseWorker):
    def __init__(self, *args, **kwargs):
        super(Worker, self).__init__()
        self.signals = WorkerSignals()
        self.args = args
        self.kwargs = kwargs
        self.plugin = None
        self.cmds = None
        self.ctx = None

    @Slot()
    def run(self):
        responses = []
        for item in self.cmds:
            if self.is_stopped():
                break
            try:
                response = None
                if (item["cmd"] in self.plugin.allowed_cmds
                        and (self.plugin.has_cmd(item["cmd"]) or 'force' in item)):

                    if item["cmd"] == "code_execute_file":
                        response = self.cmd_code_execute_file(item)

                    elif item["cmd"] == "code_execute":
                        response = self.cmd_code_execute(item)
                        if "silent" in item:
                            response = None

                    elif item["cmd"] == "code_execute_all":
                        response = self.cmd_code_execute_all(item)
                        if "silent" in item:
                            response = None

                    elif item["cmd"] == "ipython_execute_new":
                        response = self.cmd_ipython_execute_new(item)
                        if "silent" in item:
                            response = None

                    elif item["cmd"] == "ipython_execute":
                        response = self.cmd_ipython_execute(item)
                        if "silent" in item:
                            response = None

                    elif item["cmd"] == "ipython_kernel_restart":
                        response = self.cmd_ipython_kernel_restart(item)
                        if "silent" in item:
                            response = None

                    elif item["cmd"] == "get_python_output":
                        response = self.cmd_get_python_output(item)

                    elif item["cmd"] == "get_python_input":
                        response = self.cmd_get_python_input(item)

                    elif item["cmd"] == "clear_python_output":
                        response = self.cmd_clear_python_output(item)

                    elif item["cmd"] == "render_html_output":
                        response = self.cmd_render_html_output(item)

                    elif item["cmd"] == "get_html_output":
                        response = self.cmd_get_html_output(item)

                    if response:
                        responses.append(response)

            except Exception as e:
                responses.append(
                    self.make_response(
                        item,
                        self.throw_error(e)
                    )
                )

        # send response
        if len(responses) > 0:
            self.reply_more(responses)

    def cmd_ipython_execute_new(self, item: dict) -> dict:
        """
        Execute code in IPython interpreter (new kernel)

        :param item: command item
        :return: response item
        """
        try:
            result = self.plugin.runner.ipython_execute_new(
                ctx=self.ctx,
                item=item,
                request=self.from_request(item),
            )
        except Exception as e:
            result = self.throw_error(e)

        extra = self.prepare_extra(item, result)
        return self.make_response(item, result, extra=extra)

    def cmd_ipython_execute(self, item: dict) -> dict:
        """
        Execute code in IPython interpreter (current kernel)

        :param item: command item
        :return: response item
        """
        try:
            result = self.plugin.runner.ipython_execute(
                ctx=self.ctx,
                item=item,
                request=self.from_request(item),
            )
        except Exception as e:
            result = self.throw_error(e)

        extra = self.prepare_extra(item, result)
        return self.make_response(item, result, extra=extra)

    def cmd_ipython_kernel_restart(self, item: dict) -> dict:
        """
        Restart current IPython kernel

        :param item: command item
        :return: response item
        """
        try:
            result = self.plugin.runner.ipython_kernel_restart(
                ctx=self.ctx,
                item=item,
                request=self.from_request(item),
            )
        except Exception as e:
            result = self.throw_error(e)

        extra = {
            'plugin': "cmd_code_interpreter",
        }
        return self.make_response(item, result, extra=extra)

    def cmd_code_execute_file(self, item: dict) -> dict:
        """
        Execute code command from existing file

        :param item: command item
        :return: response item
        """
        request = self.from_request(item)
        try:
            if not self.plugin.runner.is_sandbox():
                result = self.plugin.runner.code_execute_file_host(
                    ctx=self.ctx,
                    item=item,
                    request=request,
                )
            else:
                result = self.plugin.runner.code_execute_file_sandbox(
                    ctx=self.ctx,
                    item=item,
                    request=request,
                )
        except Exception as e:
            result = self.throw_error(e)

        extra = self.prepare_extra(item, result)
        return self.make_response(item, result, extra=extra)

    def cmd_code_execute(self, item: dict) -> dict:
        """
        Execute code command

        :param item: command item
        :return: response item
        """
        request = self.from_request(item)
        try:
            if not self.plugin.runner.is_sandbox():
                result = self.plugin.runner.code_execute_host(
                    ctx=self.ctx,
                    item=item,
                    request=request,
                )
            else:
                result = self.plugin.runner.code_execute_sandbox(
                    ctx=self.ctx,
                    item=item,
                    request=request,
                )
        except Exception as e:
            result = self.throw_error(e)

        extra = self.prepare_extra(item, result)
        return self.make_response(item, result, extra=extra)

    def cmd_code_execute_all(self, item: dict) -> dict:
        """
        Execute all code command

        :param item: command item
        :return: response item
        """
        request = self.from_request(item)
        try:
            if not self.plugin.runner.is_sandbox():
                result = self.plugin.runner.code_execute_host(
                    ctx=self.ctx,
                    item=item,
                    request=request,
                    all=True,
                )
            else:
                result = self.plugin.runner.code_execute_sandbox(
                    ctx=self.ctx,
                    item=item,
                    request=request,
                    all=True,
                )
        except Exception as e:
            result = self.throw_error(e)

        extra = self.prepare_extra(item, result)
        return self.make_response(item, result, extra=extra)

    def cmd_get_python_output(self, item: dict) -> dict:
        """
        Get python output

        :param item: command item
        :return: response item
        """
        try:
            result = self.plugin.window.tools.get("interpreter").get_current_output()
        except Exception as e:
            result = self.throw_error(e)

        extra = self.prepare_extra(item, result)
        return self.make_response(item, result, extra=extra)

    def cmd_get_python_input(self, item: dict) -> dict:
        """
        Get python input (edit code)

        :param item: command item
        :return: response item
        """
        try:
            result = self.plugin.window.tools.get("interpreter").get_current_history()
        except Exception as e:
            result = self.throw_error(e)

        extra = self.prepare_extra(item, result)
        return self.make_response(item, result, extra=extra)

    def cmd_clear_python_output(self, item: dict) -> dict:
        """
        Clear python output

        :param item: command item
        :return: response item
        """
        try:
            self.signals.clear.emit()
            result = "OK"
        except Exception as e:
            result = self.throw_error(e)
        return self.make_response(item, result)

    def cmd_render_html_output(self, item: dict) -> dict:
        """
        Show output in HTML canvas

        :param item: command item
        :return: response item
        """
        try:
            if self.has_param(item, "html"):
                self.plugin.runner.send_html_output(self.get_param(item, "html")) # handle in main thread
            result = "OK"
        except Exception as e:
            result = self.throw_error(e)
        return self.make_response(item, result)

    def cmd_get_html_output(self, item: dict) -> dict:
        """
        Get HTML canvas output

        :param item: command item
        :return: response item
        """
        try:
            result = self.plugin.window.tools.get("html_canvas").get_output()
        except Exception as e:
            result = self.throw_error(e)

        extra = self.prepare_extra(item, result)
        return self.make_response(item, result, extra=extra)

    def prepare_extra(self, item: dict, result: dict) -> dict:
        """
        Prepare extra data for response

        :param item: command item
        :param result: response data
        :return: extra data
        """
        cmd = item["cmd"]
        extra = {
            'plugin': "cmd_code_interpreter",
            'cmd': cmd,
            'code': {}
        }
        lang = "python"
        if cmd in ["render_html_output", "get_html_output"]:
            lang = "html"
        elif cmd in ["sys_exec"]:
            lang = "bash"
        if "params" in item and "code" in item["params"]:
            extra["code"]["input"] = {}
            extra["code"]["input"]["lang"] = lang
            extra["code"]["input"]["content"] = str(item["params"]["code"])
        if "result" in result:
            extra["code"]["output"] = {}
            extra["code"]["output"]["lang"] = lang
            extra["code"]["output"]["content"] = str(result["result"])
        if "context" in result:
            extra["context"] = str(result["context"])
        return extra





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 22:00:00                  #
# ================================================== #

import copy
from typing import Optional, Any, Dict, List

from PySide6.QtCore import QObject, Slot

from pygpt_net.core.bridge.context import BridgeContext
from pygpt_net.core.events import Event, KernelEvent
from pygpt_net.item.ctx import CtxItem
from pygpt_net.utils import trans


class BasePlugin(QObject):
    def __init__(self, *args, **kwargs):
        super(BasePlugin, self).__init__()
        self.window = kwargs.get('window', None)
        self.id = ""
        self.name = ""
        self.type = []  # Types:
        # audio.input, audio.output, image.input, image.output, schedule, text.input, text.output, vision
        self.description = ""
        self.prefix = "Plugin"
        self.urls = {}
        self.options = {}
        self.initial_options = {}
        self.allowed_cmds = []
        self.tabs = {}
        self.parent = None
        self.enabled = False
        self.use_locale = False
        self.order = 0

    def setup(self) -> Dict[str, Any]:
        """
        Return available config options

        :return: config options
        """
        return self.options

    def add_option(
            self,
            name: str,
            type: str,
            **kwargs
    ) -> Dict[str, Any]:
        """
        Add plugin configuration option

        :param name: option name (ID, key)
        :param type: option type (text, textarea, bool, int, float, dict, combo)
        :param kwargs: additional keyword arguments for option properties
        :return: added option config dict
        """
        defaults = {
            "value": None,
            "label": "",
            "description": "",
            "tooltip": None,
            "min": None,
            "max": None,
            "multiplier": 1,
            "step": 1,
            "slider": False,
            "keys": None,
            "advanced": False,
            "secret": False,
            "persist": False,
            "urls": None,
            "use": None,
        }
        option = {**defaults, **kwargs}
        option['tooltip'] = option['tooltip'] or option['description']
        option["id"] = name
        option["type"] = type
        self.options[name] = option
        return option

    def add_cmd(
            self,
            cmd: str,
            **kwargs
    ) -> Dict[str, Any]:
        """
        Add plugin command

        :param cmd: command name
        :param kwargs: additional keyword arguments for command properties
        :return: added option config dict
        """
        cmd_syntax = {
            "instruction": "",  # instruction for model
            "params": {},  # parameters
            "enabled": True,  # enabled
        }
        if "instruction" in kwargs and isinstance(kwargs.get("instruction"), str):
            cmd_syntax["instruction"] = kwargs.get("instruction")
            kwargs.pop("instruction")
        if "params" in kwargs and isinstance(kwargs.get("params"), list):
            cmd_syntax["params"] = kwargs.get("params")
            kwargs.pop("params")
        if "enabled" in kwargs and isinstance(kwargs.get("enabled"), bool):
            cmd_syntax["enabled"] = kwargs.get("enabled")
            kwargs.pop("enabled")

        name = "cmd." + cmd
        kwargs["cmd"] = cmd
        kwargs["value"] = cmd_syntax

        # static keys
        kwargs["params_keys"] = {
            "name": "text",
            "type": {
                "type": "combo",
                "use": "var_types",
                "keys": {},
            },
            "description": "text",
            "required": "bool",
        }

        return self.add_option(name, "cmd", **kwargs)

    def has_cmd(
            self,
            cmd: str
    ) -> bool:
        """
        Check if command exists

        :param cmd: command name
        :return: True if exists
        """
        key = "cmd." + cmd
        if key in self.options:
            if "value" in self.options[key] and "enabled" in self.options[key]["value"]:
                return self.options[key]["value"]["enabled"]
        return False

    def cmd_allowed(
            self,
            cmd: str
    ) -> bool:
        """
        Check if command allowed

        :param cmd: command name
        :return: True if allowed
        """
        if cmd in self.allowed_cmds:
            return True
        return False

    def cmd_exe(self) -> bool:
        """
        Check if command enabled

        :return: True if enabled
        """
        return self.window.core.config.get("cmd")

    def get_cmd(
            self,
            cmd: str
    ) -> Dict[str, Any]:
        """
        Return command

        :param cmd: command name
        :return: command dict
        """
        key = "cmd." + cmd
        if key in self.options:
            data = copy.deepcopy(self.options[key]["value"]) # make copy to prevent changes in original data
            data = {"cmd": cmd, **data}
            return data

    def has_option(
            self,
            name: str
    ) -> bool:
        """
        Check if option exists

        :param name: option name
        :return: True if exists
        """
        return name in self.options

    def get_option(
            self,
            name: str
    ) -> Dict[str, Any]:
        """
        Return option

        :param name: option name
        :return: option dict
        """
        if self.has_option(name):
            return self.options[name]

    def get_option_value(
            self,
            name: str
    ) -> Any:
        """
        Return option value

        :param name: option name
        :return: option value
        """
        if self.has_option(name):
            return self.options[name]["value"]

    def attach(self, window):
        """
        Attach window to plugin

        :param window: Window instance
        """
        self.window = window

    def handle(
            self,
            event: Event,
            *args,
            **kwargs
    ):
        """
        Handle event

        :param event: event name
        :param args: arguments
        :param kwargs: keyword arguments
        """
        return

    def on_update(self, *args, **kwargs):
        """
        Called on update

        :param args: arguments
        :param kwargs: keyword arguments
        """
        return

    def on_post_update(self, *args, **kwargs):
        """
        Called on post-update

        :param args: arguments
        :param kwargs: keyword arguments
        """
        return

    def trans(
            self,
            text: Optional[str] = None
    ) -> str:
        """
        Translate text using plugin domain

        :param text: text to translate
        :return: translated text
        """
        if text is None:
            return ""
        domain = 'plugin.{}'.format(self.id)
        return trans(text, False, domain)

    def error(self, err: Any):
        """
        Send error message to logger and alert dialog

        :param err: error message
        """
        self.window.core.debug.log(err)
        msg = self.window.core.debug.parse_alert(err)
        self.window.ui.dialogs.alert("{}: {}".format(self.name, msg))

    def debug(self, data: Any):
        """
        Send debug message to logger window

        :param data: data to send
        """
        self.window.core.debug.info(data)

    def reply(
            self,
            response: dict,
            ctx: Optional[CtxItem] = None,
            extra_data: Optional[dict] = None
    ):
        """
        Send reply from plugin (command response)

        :param response: response
        :param ctx: context (CtxItem)
        :param extra_data: extra data
        """
        self.handle_finished(response, ctx, extra_data)

    def is_native_cmd(self) -> bool:
        """
        Check if native API command execution is enabled

        :return: True if native execution is enabled
        """
        return self.window.core.command.is_native_enabled()

    def is_log(self) -> bool:
        """
        Check if logging to console is enabled

        :return: True if log is enabled
        """
        if self.window.core.config.has("log.plugins") \
                and self.window.core.config.get("log.plugins"):
            return True
        return False

    def is_async(self, ctx: CtxItem) -> bool:
        """
        Check if async execution is allowed

        :param ctx: context (CtxItem)
        :return: True if async execution is allowed
        """
        return self.window.controller.kernel.async_allowed(ctx)

    def log(self, msg: str):
        """
        Log message to logger and console

        :param msg: message to log
        """
        msg = "[{}] {}".format(self.prefix, msg)
        self.debug(msg)
        if self.is_threaded():
            return
        self.window.update_status(msg.replace("\n", " "))
        if self.is_log():
            print(msg)

    def cmd_prepare(self, ctx: CtxItem, cmds: list):
        """
        On command run

        :param ctx: CtxItem
        :param cmds: commands dict
        """
        # set state: busy
        self.window.dispatch(KernelEvent(KernelEvent.STATE_BUSY, {
            "id": "img",
        }))

    @Slot(object, object, dict)
    def handle_finished(
            self,
            response: Dict[str, Any],
            ctx: Optional[CtxItem] = None,
            extra_data: Optional[Dict[str, Any]] = None
    ):
        """
        Handle finished response signal

        :param response: response
        :param ctx: context (CtxItem)
        :param extra_data: extra data
        """
        # send response (reply)
        if ctx is not None:
            self.prepare_reply_ctx(response, ctx)
            
            context = BridgeContext()
            context.ctx = ctx
            extra = {}
            if extra_data:
                extra = extra_data
            extra["response_type"] = "single"
            event = KernelEvent(KernelEvent.REPLY_ADD, {
                'context': context,
                'extra': extra,
            })
            self.window.dispatch(event)

    @Slot(object, object, dict)
    def handle_finished_more(
            self,
            responses: List[Dict[str, Any]],
            ctx: Optional[CtxItem] = None,
            extra_data: Optional[Dict[str, Any]] = None
    ):
        """
        Handle finished response signal

        :param responses: responses
        :param ctx: context (CtxItem)
        :param extra_data: extra data
        """
        # send multiple responses (reply)
        for response in responses:
            if ctx is not None:
                self.prepare_reply_ctx(response, ctx)

        context = BridgeContext()
        context.ctx = ctx
        extra = {}
        if extra_data:
            extra = extra_data
        extra["response_type"] = "multiple"
        event = KernelEvent(KernelEvent.REPLY_ADD, {
            'context': context,
            'extra': extra,
        })
        self.window.dispatch(event)

    def prepare_reply_ctx(
            self,
            response: Dict[str, Any],
            ctx: Optional[CtxItem] = None
    ) -> Dict[str, Any]:
        """
        Prepare reply context

        :param response: response
        :param ctx: context (CtxItem)
        :return: response dict
        """
        ignore_extra = ["request", "result", "context"]
        allow_output = ["request", "result", "context"]
        clean_response = {}
        for key in response:
            if key in allow_output or key.startswith("agent_"):
                clean_response[key] = response[key]

        ctx.results.append(clean_response)
        ctx.reply = True

        extras = {}
        for key in response:
            if key not in ignore_extra:
                extras[key] = response[key]

        # add extra data to context, into `tool_output` list
        if not isinstance(ctx.extra, dict):
            ctx.extra = {}
        if "tool_output" not in ctx.extra:
            ctx.extra["tool_output"] = []
        ctx.extra["tool_output"].append(extras)  # allow more extra data

        if "context" in response:
            if self.window.core.config.get("ctx.use_extra"):
                if ctx.extra_ctx is None:
                    ctx.extra_ctx = ""
                if ctx.extra_ctx != "":
                    ctx.extra_ctx += "\n\n"
                ctx.extra_ctx += response["context"]  # allow more context data
                response["result"] = "OK"
            else:
                del response["context"]

        return response

    @Slot(object)
    def handle_status(self, data: str):
        """
        Handle thread status msg signal

        :param data: status message
        """
        if self.is_threaded():
            return
        self.window.update_status(str(data))

    @Slot(object)
    def handle_error(self, err: Any):
        """
        Handle thread error signal

        :param err: error message
        """
        self.error(err)

    @Slot(object)
    def handle_debug(self, msg: Any):
        """
        Handle debug message signal

        :param msg: debug message
        """
        self.debug(msg)

    @Slot(object)
    def handle_log(self, msg: str):
        """
        Handle log message signal

        :param msg: log message
        """
        if self.is_threaded():
            return
        self.log(msg)

    def is_threaded(self) -> bool:
        """
        Check if plugin is threaded

        :return: True if threaded
        """
        return self.window.controller.kernel.is_threaded()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 08:00:00                  #
# ================================================== #

from typing import Optional, List

from pygpt_net.item.ctx import CtxItem, CtxMeta
from pygpt_net.core.tabs.tab import Tab

class Output:
    def __init__(self, window=None):
        """
        Context output mapping

        :param window: Window
        """
        self.window = window
        self.mapping = {}  # [column_idx => [pid => meta.id]]
        self.last_pids = {}  # meta.id => pid
        self.last_pid = 0 # last used PID
        self.initialized = False

    def init(self, force: bool = False):
        """
        Initialize mappings

        :param force: Force reinitialize
        """
        if self.initialized and not force:
            return
        self.clear()
        for n in range(0, self.window.core.tabs.NUM_COLS):
            self.mapping[n] = {}
            self.last_pids[n] = {}
        self.initialized = True

    def store(self, meta: CtxMeta) -> int:
        """
        Store meta in mapping

        :param meta: Meta
        :return: PID
        """
        self.init()
        pid = self.window.core.tabs.get_active_pid()
        tab = self.window.core.tabs.get_tab_by_pid(pid)
        if tab is None or tab.type != Tab.TAB_CHAT:
            return 0
        col_idx = tab.column_idx
        self.mapping[col_idx][pid] = meta.id
        self.last_pids[col_idx][meta.id] = pid
        self.last_pid = pid
        tab = self.window.core.tabs.get_tab_by_pid(pid)
        if tab is not None:
            tab.data_id = meta.id  # store meta id in tab data
        return pid

    def is_mapped(self, meta: CtxMeta) -> bool:
        """
        Check if meta is mapped

        :param meta: Meta
        :return: bool
        """
        self.init()
        for col_idx in self.mapping:
            if meta.id in self.mapping[col_idx].values():
                return True
        return False

    def get_meta(self, pid: int) -> Optional[int]:
        """
        Get meta by PID

        :param pid: PID
        :return: meta ID or None
        """
        self.init()
        for col_idx in self.mapping:
            if pid in self.mapping[col_idx]:
                return self.mapping[col_idx][pid]
        return

    def prepare_meta(self, tab: Tab) -> Optional[int]:
        """
        Get meta ID by PID

        :param tab: Tab
        :return: Meta ID or None
        """
        self.init()
        pid = tab.pid
        col_idx = tab.column_idx
        meta_id = None
        if pid in self.mapping[col_idx]:
            return self.mapping[col_idx][pid]
        if meta_id is None:
            meta_id = tab.data_id
            if meta_id is not None:
                self.mapping[col_idx][pid] = meta_id
                self.last_pids[col_idx][meta_id] = pid
                self.last_pid = pid
        return meta_id

    def get_mapped(self, meta: CtxMeta) -> Optional[int]:
        """
        Get PID by meta

        :param meta: Meta
        :return: PID or None
        """
        self.init()
        all = []
        pid = None
        active_pid = self.window.core.tabs.get_active_pid()
        tab = self.window.core.tabs.get_tab_by_pid(active_pid)
        if tab is None:
            return None
        col_idx = tab.column_idx
        for pid, meta_id in self.mapping[col_idx].items():
            if meta_id == meta.id:
                all.append(pid)
        for pid in all:
            if pid == active_pid:  # at first check for current tab
                return pid
        return pid

    def is_empty(self) -> bool:
        """
        Check if mapping is empty

        :return: bool
        """
        self.init()
        active_pid = self.window.core.tabs.get_active_pid()
        for col_idx in list(self.mapping.keys()):
            if active_pid in self.mapping[col_idx]:
                return False
        return True

    def get_pid(self, meta: Optional[CtxMeta] = None) -> Optional[int]:
        """
        Get PID by meta

        :param meta: Meta
        :return: PID
        """
        self.init()
        active_pid = self.window.core.tabs.get_active_pid()
        if meta is None:
            return None
        if self.is_mapped(meta):
            if active_pid == self.get_mapped(meta):  # if loaded in current tab
                return active_pid  # return current
        return self.store(meta)

    def clear(self):
        """
        Clear mapping
        """
        self.mapping = {}
        self.last_pids = {}
        self.last_pid = 0
        self.initialized = False

    def get_current(self, meta: Optional[CtxMeta] = None):
        """
        Get current output node by meta

        :param meta: Meta
        :return: Node
        """
        pid = self.get_pid(meta)
        if pid in self.window.ui.nodes['output']:
            return self.window.ui.nodes['output'][pid]
        else:
            for pid in self.window.ui.nodes['output']:
                return self.window.ui.nodes['output'][pid]  # get first available

    def get_current_plain(self, meta: Optional[CtxMeta] = None):
        """
        Get current output plain node by meta

        :param meta: Meta
        :return: Node
        """
        pid = self.get_pid(meta)
        if pid in self.window.ui.nodes['output_plain']:
            return self.window.ui.nodes['output_plain'][pid]
        else:
            for pid in self.window.ui.nodes['output_plain']:
                return self.window.ui.nodes['output_plain'][pid]

    def get_by_pid(self, pid: Optional[int] = None):
        """
        Get output node by PID

        :param pid: PID
        :return: Node widget
        """
        if pid in self.window.ui.nodes['output']:
            return self.window.ui.nodes['output'][pid]
        else:
            for pid in self.window.ui.nodes['output']:
                return self.window.ui.nodes['output'][pid]  # get first available

    def get_by_pid_plain(self, pid: Optional[int] = None):
        """
        Get output plain node by PID

        :param pid: PID
        :return: Node widget
        """
        if pid in self.window.ui.nodes['output_plain']:
            return self.window.ui.nodes['output_plain'][pid]
        else:
            for pid in self.window.ui.nodes['output_plain']:
                return self.window.ui.nodes['output_plain'][pid]

    def get_all(self) -> List:
        """
        Get all output nodes

        :return: List of nodes
        """
        nodes = []
        for node in self.window.ui.nodes['output'].values():
            nodes.append(node)
        return nodes

    def get_all_plain(self) -> List:
        """
        Get all output plain nodes

        :return: List of nodes
        """
        nodes = []
        for node in self.window.ui.nodes['output_plain'].values():
            nodes.append(node)
        return nodes


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 22:00:00                  #
# ================================================== #

from typing import Optional, Dict

from PySide6.QtCore import QObject
from PySide6.QtGui import QAction
from PySide6.QtWidgets import QWidget

from pygpt_net.core.tabs.tab import Tab
from pygpt_net.ui.widget.dialog.base import BaseDialog

class BaseTool(QObject):
    def __init__(self, *args, **kwargs):
        """
        Base Tool

        :param window: Window instance
        :param args: arguments
        :param kwargs: keyword arguments
        """
        super(BaseTool, self).__init__()
        self.window = None
        self.id = ""
        self.has_tab = False
        self.tab_title = ""
        self.tab_icon = ":/icons/build.svg"

    def setup(self):
        """Setup tool"""
        pass

    def post_setup(self):
        """Post-setup (window), after plugins are loaded"""
        pass

    def on_update(self):
        """On app main loop update"""
        pass

    def on_post_update(self):
        """On app main loop update"""
        pass

    def on_exit(self):
        """On app exit"""
        pass

    def on_reload(self):
        """On app profile reload"""
        pass

    def attach(self, window):
        """
        Attach window to plugin

        :param window: Window instance
        """
        self.window = window

    def setup_menu(self) -> Dict[str, QAction]:
        """
        Setup main menu

        :return dict with menu actions
        """
        return {}

    def setup_dialogs(self):
        """Setup dialogs (static)"""
        pass

    def setup_theme(self):
        """Setup theme"""
        pass

    def get_instance(
            self,
            type_id: str,
            dialog_id: Optional[str] = None
    ) -> Optional[BaseDialog]:
        """
        Spawn and return dialog instance

        :param type_id: dialog instance ID
        :param dialog_id: dialog instance ID
        """
        return None

    def as_tab(self, tab: Tab) -> Optional[QWidget]:
        """
        Spawn and return tab instance

        :param tab: Parent tab instance
        :return: Tab widget instance
        """
        return None

    def get_lang_mappings(self) -> Dict[str, Dict]:
        """
        Get language mappings

        :return: dict with language mappings
        """
        return {}


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.18 19:00:00                  #
# ================================================== #

from pygpt_net.plugin.base.config import BaseConfig, BasePlugin


class Config(BaseConfig):
    def __init__(self, plugin: BasePlugin = None, *args, **kwargs):
        super(Config, self).__init__(plugin)
        self.plugin = plugin

    def from_defaults(self, plugin: BasePlugin = None):
        """
        Set default options for plugin

        :param plugin: plugin instance
        """
        plugin.add_option(
            "hour",
            type="bool",
            value=True,
            label="Append time",
            description="If enabled, current time will be appended to system prompt.",
            tooltip="Hour will be appended to system prompt.",
        )
        plugin.add_option(
            "date",
            type="bool",
            value=True,
            label="Append date",
            description="If enabled, current date will be appended to system prompt.",
            tooltip="Date will be appended to system prompt.",
        )

        desc = "Template to append to system prompt.\n" \
               "Placeholder {time} will be replaced with current date and time in real-time. "
        tooltip = "Text to append to system prompt."
        plugin.add_option(
            "tpl",
            type="textarea",
            value="Current time is {time}.",
            label="Template",
            description=desc,
            tooltip=tooltip,
        )
        # commands
        plugin.add_cmd(
            "get_time",
            instruction="get current time and date",
            params=[],
            enabled=True,
            description="Enable: Get current time and date.",
        )


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 08:00:00                  #
# ================================================== #

import datetime
from typing import Dict, Any

from PySide6.QtWidgets import QHBoxLayout, QLabel, QMenu, QLayout

from pygpt_net.ui.widget.element.button import ContextMenuButton
from pygpt_net.ui.widget.element.labels import HelpLabel, TitleLabel
from pygpt_net.utils import trans


class Settings:
    def __init__(self, window=None):
        """
        Index settings controller

        :param window: Window instance
        """
        self.window = window

    def append_tabs(self) -> list:
        """
        Return additional tabs list

        :return: list of tab IDs
        """
        return ["update"]

    def append(
            self,
            content: Dict[str, QLayout],
            widgets: Dict[str, Any],
            options: Dict[str, Any],
    ):
        """
        Append extra settings to settings dialog (section: llama-index)

        :param content: settings widgets layout
        :param widgets: settings widgets
        :param options: settings options config fields
        """
        btns = QHBoxLayout()
        self.window.ui.nodes['idx.btn.db.index_all'] = \
            ContextMenuButton(trans('settings.llama.extra.btn.idx_db_all'))  # index DB (all)
        self.window.ui.nodes['idx.btn.db.index_all'].action = self.idx_db_all_context_menu
        self.window.ui.nodes['idx.btn.db.index_update'] = \
            ContextMenuButton(trans('settings.llama.extra.btn.idx_db_update'))  # index DB (only update)
        self.window.ui.nodes['idx.btn.db.index_update'].action = self.idx_db_update_context_menu
        self.window.ui.nodes['idx.btn.db.index_files'] = \
            ContextMenuButton(trans('settings.llama.extra.btn.idx_files_all'))  # index files (data)
        self.window.ui.nodes['idx.btn.db.index_files'].action = self.idx_data_context_menu

        self.window.ui.nodes['idx.api.warning'] = TitleLabel(trans('settings.llama.extra.api.warning'))
        self.window.ui.nodes['idx.api.warning'].setWordWrap(True)

        self.window.ui.nodes['idx.db.last_updated'] = QLabel("")
        self.update_text_last_updated()
        btns.addWidget(self.window.ui.nodes['idx.btn.db.index_all'])
        btns.addWidget(self.window.ui.nodes['idx.btn.db.index_update'])
        btns.addWidget(self.window.ui.nodes['idx.btn.db.index_files'])

        # offline loaders
        self.window.ui.nodes['idx.db.settings.loaders'] = QLabel("")
        self.update_text_loaders()
        self.window.ui.nodes['idx.db.settings.loaders'].setWordWrap(True)
        self.window.ui.nodes['idx.db.settings.legend.head'] = TitleLabel(trans('settings.llama.extra.btn.idx_head'))
        self.window.ui.nodes['idx.db.settings.legend'] = HelpLabel(trans('settings.llama.extra.legend'), self.window)
        self.window.ui.nodes['idx.db.settings.legend'].setWordWrap(True)

        if "data_loaders" in content:
            content["data_loaders"].addWidget(self.window.ui.nodes['idx.db.settings.loaders'])

        if "update" in content:
            content["update"].addWidget(self.window.ui.nodes['idx.db.settings.legend.head'])
            content["update"].addLayout(btns)
            content["update"].addWidget(self.window.ui.nodes['idx.db.settings.legend'])
            content["update"].addWidget(self.window.ui.nodes['idx.db.last_updated'])
            content["update"].addWidget(self.window.ui.nodes['idx.api.warning'])

    def update_text_last_updated(self):
        """Update last updated text"""
        last_str = trans('settings.llama.extra.db.never')
        if self.window.core.config.has('llama.idx.db.last'):
            last_ts = int(self.window.core.config.get('llama.idx.db.last'))
            if last_ts > 0:
                # convert timestamp to datetime
                last_str = datetime.datetime.fromtimestamp(last_ts).strftime('%Y-%m-%d %H:%M:%S')

        txt = trans('idx.last') + ": " + last_str
        self.window.ui.nodes['idx.db.last_updated'].setText(txt)

    def update_text_loaders(self):
        """Update text loaders list"""
        list_file = []
        list_web = []
        list_file.append("- Txt/raw files (txt)")
        providers = self.window.core.idx.indexing.get_data_providers()
        for id in providers:
            loader = providers[id]
            if "file" in loader.type:
                list_file.append("- " + loader.name + " (" + ", ".join(loader.extensions) + ") - file_" + loader.id)
            if "web" in loader.type:
                list_web.append("- " + loader.name + " - web_" + loader.id)
        list_file = sorted(list_file)
        list_web = sorted(list_web)
        files_str = "\n".join(list_file)
        web_str = "\n".join(list_web)
        info = trans('settings.llama.extra.loaders') + ":\n\nFiles:\n" + files_str + "\n\nWeb/external:\n" + web_str
        self.window.ui.nodes['idx.db.settings.loaders'].setText(info)

    def update_idx_choices(self):
        """Update index choices"""
        option = self.window.controller.settings.editor.get_option("agent.idx")
        option["keys"] = self.window.controller.config.placeholder.get_idx()
        self.window.ui.config["config"]["agent.idx"].combo.clear()
        self.window.ui.config["config"]["agent.idx"].option = option
        self.window.ui.config["config"]["agent.idx"].update()

    def idx_db_all_context_menu(self, parent, pos):
        """
        Index DB (all) btn context menu

        :param parent: parent widget (button)
        :param pos: mouse position
        """
        menu = QMenu(parent)
        idxs = self.window.core.config.get('llama.idx.list')
        if len(idxs) > 0:
            for idx in idxs:
                id = idx['id']
                name = idx['name'] + " (" + idx['id'] + ")"
                action = menu.addAction("IDX: " + name)
                action.triggered.connect(
                    lambda checked=False,
                           id=id: self.window.controller.idx.indexer.index_ctx_from_ts(id, 0)
                )
        menu.exec_(parent.mapToGlobal(pos))

    def idx_db_update_context_menu(self, parent, pos):
        """
        Index DB (update) btn context menu

        :param parent: parent widget (button)
        :param pos: mouse  position
        """
        menu = QMenu(parent)
        idxs = self.window.core.config.get('llama.idx.list')
        if len(idxs) > 0:
            for idx in idxs:
                id = idx['id']
                name = idx['name'] + " (" + idx['id'] + ")"
                action = menu.addAction("IDX: " + name)
                action.triggered.connect(
                    lambda checked=False,
                           id=id: self.window.controller.idx.indexer.index_ctx_current(id))
        menu.exec_(parent.mapToGlobal(pos))

    def idx_data_context_menu(self, parent, pos):
        """
        Index files (data) btn context menu

        :param parent: parent widget (button)
        :param pos: mouse  position
        """
        menu = QMenu(parent)
        idxs = self.window.core.config.get('llama.idx.list')
        if len(idxs) > 0:
            for idx in idxs:
                id = idx['id']
                name = idx['name'] + " (" + idx['id'] + ")"
                action = menu.addAction("IDX: " + name)
                action.triggered.connect(
                    lambda checked=False,
                           id=id: self.window.controller.idx.indexer.index_all_files(id))
        menu.exec_(parent.mapToGlobal(pos))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.01.19 02:00:00                  #
# ================================================== #

import os
import sys
import argparse
from logging import ERROR, WARNING, INFO, DEBUG

from PySide6 import QtCore
from PySide6.QtCore import QCoreApplication, Qt
from PySide6.QtGui import QScreen
from PySide6.QtWidgets import QApplication

from pygpt_net.core.events import AppEvent
from pygpt_net.core.access.shortcuts import GlobalShortcutFilter
from pygpt_net.core.debug import Debug
from pygpt_net.core.platforms import Platforms
from pygpt_net.provider.agents.base import BaseAgent
from pygpt_net.tools import BaseTool
from pygpt_net.ui.main import MainWindow
from pygpt_net.plugin.base.plugin import BasePlugin
from pygpt_net.provider.llms.base import BaseLLM
from pygpt_net.provider.loaders.base import BaseLoader
from pygpt_net.provider.vector_stores.base import BaseStore
from pygpt_net.provider.audio_input.base import BaseProvider as BaseAudioInput
from pygpt_net.provider.audio_output.base import BaseProvider as BaseAudioOutput
from pygpt_net.provider.web.base import BaseProvider as BaseWeb


class Launcher:
    def __init__(self):
        """Launcher"""
        self.app = None
        self.window = None
        self.debug = False
        self.force_legacy = False
        self.force_disable_gpu = False
        self.shortcut_filter = None
        self.workdir = None

    def setup(self) -> dict:
        """
        Setup launcher

        :return: dict with launcher arguments
        """
        parser = argparse.ArgumentParser()
        parser.add_argument(
            "-d",
            "--debug",
            required=False,
            help="debug mode (0=disabled, 1=info, 2=debug)",
        )
        parser.add_argument(
            "-l",
            "--legacy",
            required=False,
            help="force enable legacy mode (0=disabled, 1=enable)",
        )
        parser.add_argument(
            "-n",
            "--disable-gpu",
            required=False,
            help="force disable OpenGL (1=disabled, 0=enabled)",
        )
        parser.add_argument(
            "-w",
            "--workdir",
            required=False,
            help="force set workdir",
        )
        args = vars(parser.parse_args())

        # set log level [ERROR|WARNING|INFO|DEBUG]
        if "debug" in args and args["debug"] == "1":
            print("** Debug mode enabled (1=INFO)")
            Debug.init(INFO)
            self.debug = True
        elif "debug" in args and args["debug"] == "2":
            print("** Debug mode enabled (2=DEBUG)")
            Debug.init(DEBUG)
            self.debug = True
        else:
            Debug.init(ERROR)  # default log level

        # force legacy mode
        if "legacy" in args and args["legacy"] == "1":
            print("** Force legacy mode enabled")
            self.force_legacy = True

        # force disable GPU
        if "disable-gpu" in args and args["disable-gpu"] == "1":
            print("** Force disable GPU enabled")
            self.force_disable_gpu = True

        # force set workdir
        if "workdir" in args and args["workdir"] is not None:
            # set as environment variable
            os.environ["PYGPT_WORKDIR"] = args["workdir"]

        return args

    def init(self):
        """Initialize app"""
        args = self.setup()
        Platforms.prepare()  # setup platform specific options
        QCoreApplication.setAttribute(Qt.AA_ShareOpenGLContexts)
        self.app = QApplication(sys.argv)
        self.app.setAttribute(QtCore.Qt.AA_DontUseNativeMenuBar)
        self.window = MainWindow(self.app, args=args)
        self.shortcut_filter = GlobalShortcutFilter(self.window)

    def add_plugin(self, plugin: BasePlugin):
        """
        Register plugin

        :param plugin: plugin instance
        """
        if not isinstance(plugin, BasePlugin):
            raise TypeError(
                "Plugin must be instance of: "
                "pygpt_net.plugin.base.BasePlugin"
            )
        self.window.add_plugin(plugin)
        if self.debug:
            print("Loaded plugin: {} ({})".format(plugin.id, plugin.__class__.__name__))

    def add_llm(self, llm: BaseLLM):
        """
        Register LLM provider

        :param llm: LLM provider instance
        """
        if not isinstance(llm, BaseLLM):
            raise TypeError(
                "LLM provider must be instance of: "
                "pygpt_net.provider.llms.base.BaseLLM"
            )
        self.window.add_llm(llm)
        if self.debug:
            print("Loaded LLM: {} ({})".format(llm.id, llm.__class__.__name__))

    def add_vector_store(self, store: BaseStore):
        """
        Register vector store provider

        :param store: Vector store provider instance
        """
        if not isinstance(store, BaseStore):
            raise TypeError(
                "Vector store provider must be instance of: "
                "pygpt_net.provider.vector_stores.base.BaseStore"
            )
        self.window.add_vector_store(store)
        if self.debug:
            print("Loaded vector store: {} ({})".format(store.id, store.__class__.__name__))

    def add_loader(self, loader: BaseLoader):
        """
        Register data loader

        :param loader: Data loader instance
        """
        if not isinstance(loader, BaseLoader):
            raise TypeError(
                "Data loader must be instance of: "
                "pygpt_net.provider.loaders.base.BaseLoader"
            )
        self.window.add_loader(loader)
        if self.debug:
            print("Loaded data loader: {} ({})".format(loader.id, loader.__class__.__name__))

    def add_audio_input(self, audio: BaseAudioInput):
        """
        Register audio input provider

        :param audio: Audio input provider instance
        """
        if not isinstance(audio, BaseAudioInput):
            raise TypeError(
                "Audio input provider must be instance of: "
                "pygpt_net.provider.audio_input.base.BaseProvider"
            )
        self.window.add_audio_input(audio)
        if self.debug:
            print("Loaded audio input: {} ({})".format(audio.id, audio.__class__.__name__))

    def add_audio_output(self, audio: BaseAudioOutput):
        """
        Register audio output provider

        :param audio: Audio output provider instance
        """
        if not isinstance(audio, BaseAudioOutput):
            raise TypeError(
                "Audio output provider must be instance of: "
                "pygpt_net.provider.audio_output.base.BaseProvider"
            )
        self.window.add_audio_output(audio)
        if self.debug:
            print("Loaded audio output: {} ({})".format(audio.id, audio.__class__.__name__))

    def add_web(self, provider: BaseWeb):
        """
        Register web provider

        :param provider: Web provider instance
        """
        if not isinstance(provider, BaseWeb):
            raise TypeError(
                "Web provider must be instance of: "
                "pygpt_net.provider.web.base.BaseProvider"
            )
        self.window.add_web(provider)
        if self.debug:
            print("Loaded web provider: {} ({})".format(provider.id, provider.__class__.__name__))

    def add_tool(self, tool: BaseTool):
        """
        Register tool

        :param tool: tool instance
        """
        if not isinstance(tool, BaseTool):
            raise TypeError(
                "Tool must be instance of: "
                "pygpt_net.tools.base.BaseTool"
            )
        self.window.add_tool(tool)
        if self.debug:
            print("Loaded tool: {} ({})".format(tool.id, tool.__class__.__name__))

    def add_agent(self, agent: BaseAgent):
        """
        Register agent (LlamaIndex agent)

        :param agent: Agent instance
        """
        if not isinstance(agent, BaseAgent):
            raise TypeError(
                "Agent must be instance of: "
                "pygpt_net.provider.agents.base.BaseAgent"
            )
        self.window.add_agent(agent)
        if self.debug:
            print("Loaded agent: {} ({})".format(agent.id, agent.__class__.__name__))

    def run(self):
        """Run app"""
        self.window.setup()
        geometry = self.window.screen().availableGeometry()
        pos = QScreen.availableGeometry(QApplication.primaryScreen()).topLeft()
        margin = 100
        self.window.resize(geometry.width() - margin, geometry.height() - margin)
        self.window.show()
        self.window.move(pos)
        self.window.post_setup()
        self.app.setWindowIcon(self.window.ui.get_app_icon())
        self.window.ui.tray.setup(self.app)
        self.window.controller.after_setup()
        self.window.dispatch(AppEvent(AppEvent.APP_STARTED))  # app event
        self.window.setup_global_shortcuts()
        sys.exit(self.app.exec())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.26 04:00:00                  #
# ================================================== #

from pygpt_net.utils import trans


class PidData():

    def __init__(self, pid, meta=None):
        """Pid Data"""
        self.pid = pid
        self.meta = meta
        self.images_appended = []
        self.urls_appended = []
        self.files_appended = []
        self.buffer = ""  # stream buffer
        self.is_cmd = False
        self.html = ""  # html buffer
        self.document = ""
        self.initialized = False
        self.loaded = False  # page loaded
        self.item = None  # current item
        self.use_buffer = False  # use html buffer
        self.name_user = trans("chat.name.user")
        self.name_bot = trans("chat.name.bot")
        self.last_time_called = 0
        self.cooldown = 1 / 6  # max chunks to parse per second
        self.throttling_min_chars = 5000  # min chunk chars to activate cooldown



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 00:00:00                  #
# ================================================== #

from typing import List

from llama_index.core.llms import ChatMessage, MessageRole
from pygpt_net.core.bridge.context import BridgeContext


class Memory:
    def __init__(self, window=None):
        """
        Agent memory

        :param window: Window instance
        """
        self.window = window

    def prepare(
            self,
            context: BridgeContext
    ) -> List[ChatMessage]:
        """
        Prepare history for agent

        :param context: BridgeContext
        :return: list with history items
        """
        messages = []

        input_prompt = context.ctx.input
        history = context.history
        mode = context.mode
        model = context.model
        model_id = model.id

        # tokens config
        used_tokens = self.window.core.tokens.from_user(input_prompt, "")  # threshold and extra included
        max_tokens = self.window.core.config.get('max_total_tokens')
        model_ctx = self.window.core.models.get_num_ctx(model_id)

        # fit to max model tokens
        if max_tokens > model_ctx:
            max_tokens = model_ctx

        if self.window.core.config.get('use_context'):
            items = self.window.core.ctx.get_history(
                history,
                model_id,
                mode,
                used_tokens,
                max_tokens,
            )
            # extract only input and output messages, skip step messages
            for item in items:
                if item.extra is not None and type(item.extra) == dict:
                    # agent input
                    if item.extra.get("agent_input", False):
                        if item.final_input is not None and item.final_input != "":
                            messages.append(ChatMessage(
                                role=MessageRole.USER,
                                content=item.final_input
                            ))
                    # agent output
                    if item.extra.get("agent_output", False):
                        if item.final_output is not None and item.final_output != "":
                            messages.append(ChatMessage(
                                role=MessageRole.ASSISTANT,
                                content=item.final_output
                            ))

        return messages


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.24 22:00:00                  #
# ================================================== #

from PySide6.QtGui import QAction, QIcon

from pygpt_net.utils import trans
import pygpt_net.icons_rc

class Tools:
    def __init__(self, window=None):
        """
        Tools menu setup

        :param window: Window instance
        """
        self.window = window

    def setup(self):
        """Setup tools menu"""
        # tab tools
        tab_tools = self.window.controller.tools.get_tab_tools()
        for key in tab_tools:
            self.window.ui.menu[key] = QAction(QIcon(":/icons/" + tab_tools[key][1] + ".svg"),
                                               trans("output.tab." + tab_tools[key][0]), self.window)
            self.window.ui.menu[key].triggered.connect(
                lambda checked=False, type=tab_tools[key][2] : self.window.controller.tools.open_tab(type)
            )
            self.window.ui.menu[key].setCheckable(False)

        # add menu
        self.window.ui.menu['menu.tools'] = self.window.menuBar().addMenu(trans("menu.tools"))

        # add tab tools
        for key in tab_tools:
            self.window.ui.menu['menu.tools'].addAction(self.window.ui.menu[key])

        # add custom tools
        actions = self.window.tools.setup_menu_actions()
        if len(actions) == 0:
            return

        # add separator
        self.window.ui.menu['menu.tools'].addSeparator()

        # build custom tools menu
        for key in actions:
            self.window.ui.menu[key] = actions[key]
            self.window.ui.menu['menu.tools'].addAction(self.window.ui.menu[key])

        # ------------------------------------------------- #

        # Docker images rebuild:

        # IPython container
        self.window.ui.menu['menu.tools'].addSeparator()
        self.window.ui.menu['menu.tools.ipython.rebuild'] = QAction(QIcon(":/icons/reload.svg"),
                                                                    "Rebuild IPython Docker Image", self.window)
        self.window.ui.menu['menu.tools'].addAction(self.window.ui.menu['menu.tools.ipython.rebuild'])
        self.window.ui.menu['menu.tools.ipython.rebuild'].triggered.connect(
            lambda: self.window.core.plugins.get("cmd_code_interpreter").builder.build_and_restart()
        )
        # Python Legacy container
        self.window.ui.menu['menu.tools.python_legacy.rebuild'] = QAction(QIcon(":/icons/reload.svg"),
                                                                          "Rebuild Python (Legacy) Docker Image", self.window)
        self.window.ui.menu['menu.tools'].addAction(self.window.ui.menu['menu.tools.python_legacy.rebuild'])
        self.window.ui.menu['menu.tools.python_legacy.rebuild'].triggered.connect(
            lambda: self.window.core.plugins.get("cmd_code_interpreter").docker.build_and_restart()
        )
        # System container
        self.window.ui.menu['menu.tools.system.rebuild'] = QAction(QIcon(":/icons/reload.svg"),
                                                                   "Rebuild System Sandbox Docker Image", self.window)
        self.window.ui.menu['menu.tools'].addAction(self.window.ui.menu['menu.tools.system.rebuild'])
        self.window.ui.menu['menu.tools.system.rebuild'].triggered.connect(
            lambda: self.window.core.plugins.get("cmd_system").docker.build_and_restart()
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.16 01:00:00                  #
# ================================================== #

import os
from typing import List, Dict, Any

from PySide6.QtCore import Slot, QObject

from pygpt_net.core.types import (
    MODE_ASSISTANT,
)
from pygpt_net.core.events import KernelEvent
from pygpt_net.core.bridge import BridgeContext
from pygpt_net.core.attachments.worker import AttachmentWorker
from pygpt_net.item.attachment import AttachmentItem
from pygpt_net.item.ctx import CtxMeta, CtxItem
from pygpt_net.utils import trans


class Attachment(QObject):

    MODE_FULL_CONTEXT = 'full'  # attach full context to system prompt
    MODE_QUERY_CONTEXT = 'query'  # query context only
    MODE_QUERY_CONTEXT_SUMMARY = 'summary'  # summary full context and attach summary to system prompt
    MODE_DISABLED = 'disabled'  # disabled

    def __init__(self, window=None):
        """
        Attachment controller

        :param window: Window instance
        """
        super(Attachment, self).__init__()
        self.window = window
        self.mode = self.MODE_FULL_CONTEXT
        self.uploaded = False
        self.uploaded_tab_idx = 3

    def has(self, mode: str) -> bool:
        """
        Check if has attachments

        :param mode: Work mode
        :return: True if has attachments
        """
        if self.window.core.attachments.has(mode):
            files = self.window.core.attachments.get_all(mode, only_files=True)
            for id in files:
                file = files[id]
                if self.is_allowed(file.path):
                    return True
        return False

    def setup(self):
        """Setup attachments"""
        self.mode = self.window.core.config.get("ctx.attachment.mode", self.MODE_FULL_CONTEXT)
        if self.mode == self.MODE_QUERY_CONTEXT:
            self.window.ui.nodes['input.attachments.ctx.mode.query'].setChecked(True)
        elif self.mode == self.MODE_QUERY_CONTEXT_SUMMARY:
            self.window.ui.nodes['input.attachments.ctx.mode.query_summary'].setChecked(True)
        elif self.mode == self.MODE_FULL_CONTEXT:
            self.window.ui.nodes['input.attachments.ctx.mode.full'].setChecked(True)
        elif self.mode == self.MODE_DISABLED:
            self.window.ui.nodes['input.attachments.ctx.mode.off'].setChecked(True)

    def reload(self):
        """Reload attachments"""
        self.setup()

    def handle(
            self,
            mode: str,
            text: str
    ):
        """
        Handle attachment upload

        :param mode: mode
        :param text: input prompt text
        """
        worker = AttachmentWorker(self.window)
        worker.window = self.window
        worker.meta = self.window.core.ctx.get_current_meta()
        worker.mode = mode
        worker.prompt = text
        worker.signals.error.connect(self.handle_upload_error)
        worker.signals.success.connect(self.handle_upload_success)
        self.window.threadpool.start(worker)

    def is_allowed(self, path: str) -> bool:
        """
        Check if path is allowed for indexing

        :param path: Path to file
        :return: True if allowed
        """
        allow_images = self.window.core.config.get("ctx.attachment.img", False)
        ext_allowed = self.window.core.idx.indexing.is_allowed(path)
        if ext_allowed or self.window.core.filesystem.packer.is_archive(path):
            if not self.window.core.filesystem.types.is_image(path) or allow_images:
                return True
        return False

    def upload(
            self,
            meta: CtxMeta,
            mode: str,
            prompt: str
    ) -> bool:
        """
        Upload attachments for meta

        :param meta: CtxMeta
        :param mode: str
        :param prompt: user input prompt
        :return: True if uploaded
        """
        self.uploaded = False
        auto_index = self.window.core.config.get("attachments_auto_index", False)
        attachments = self.window.core.attachments.get_all(mode, only_files=True)

        if self.is_verbose() and len(attachments) > 0:
            print("\nUploading attachments...\nWork Mode: {}".format(mode))

        for uuid in attachments:
            attachment = attachments[uuid]
            if attachment.type == AttachmentItem.TYPE_FILE:
                result = self.upload_file(
                    attachment=attachment,
                    meta=meta,
                    prompt=prompt,
                    auto_index=auto_index,
                )
                if result:
                    self.uploaded = True
            elif attachment.type == AttachmentItem.TYPE_URL:
                result = self.upload_web(
                    attachment=attachment,
                    meta=meta,
                    prompt=prompt,
                    auto_index=auto_index,
                )
                if result:
                    self.uploaded = True
        if self.uploaded:
            self.window.core.ctx.save(meta.id)  # save meta
        return self.uploaded

    def upload_file(
            self,
            attachment: AttachmentItem,
            meta: CtxMeta,
            prompt: str,
            auto_index: bool
    ) -> bool:
        """
        Upload file attachment

        :param attachment: AttachmentItem
        :param meta: CtxMeta
        :param prompt: User input prompt
        :param auto_index: Auto index
        :return: True if uploaded
        """
        uploaded = False
        if not self.is_allowed(attachment.path):
            return False
        if self.window.core.filesystem.packer.is_archive(attachment.path):
            if self.is_verbose():
                print("Unpacking archive: {}".format(attachment.path))
            tmp_path = self.window.core.filesystem.packer.unpack(attachment.path)
            if tmp_path:
                for root, dirs, files in os.walk(tmp_path):
                    for file in files:
                        path = str(os.path.join(root, file))
                        sub_attachment = AttachmentItem()
                        sub_attachment.path = path
                        sub_attachment.name = os.path.basename(path)
                        sub_attachment.consumed = False
                        path_relative = os.path.relpath(path, tmp_path)
                        if self.is_allowed(str(path)):
                            if self.is_verbose():
                                print("Uploading unpacked from archive: {}".format(path_relative))
                            item = self.window.core.attachments.context.upload(
                                meta=meta,
                                attachment=sub_attachment,
                                prompt=prompt,
                                real_path=attachment.path,
                                auto_index=auto_index,
                            )
                            if item:
                                item["path"] = os.path.basename(attachment.path) + "/" + str(path_relative)
                                item["size"] = os.path.getsize(path)
                                self.append_to_meta(meta, item)
                                uploaded = True
                                sub_attachment.consumed = True
                                attachment.consumed = True
                self.window.core.filesystem.packer.remove_tmp(tmp_path)  # clean
        else:
            item = self.window.core.attachments.context.upload(
                meta=meta,
                attachment=attachment,
                prompt=prompt,
                real_path=attachment.path,
                auto_index=auto_index,
            )
            if item:
                self.append_to_meta(meta, item)
                attachment.consumed = True  # allow for deletion
                uploaded = True
        return uploaded

    def append_to_meta(self, meta: CtxMeta, item: Dict[str, Any]):
        """
        Append item to meta

        :param meta: CtxMeta instance
        :param item: Attachment item
        """
        if meta.group:
            if meta.group.additional_ctx is None:
                meta.group.additional_ctx = []
            meta.group.additional_ctx.append(item)
            return
        if meta.additional_ctx is None:
            meta.additional_ctx = []
        meta.additional_ctx.append(item)

    def upload_web(
            self,
            attachment: AttachmentItem,
            meta: CtxMeta,
            prompt: str,
            auto_index: bool
    ) -> bool:
        """
        Upload web attachment

        :param attachment: AttachmentItem
        :param meta: CtxMeta
        :param prompt: User input prompt
        :param auto_index: Auto index
        :return: True if uploaded
        """
        return self.upload_file(attachment, meta, prompt, auto_index)

    def has_context(self, meta: CtxMeta) -> bool:
        """
        Check if has additional context for attachment

        :param meta: CtxMeta
        :return: True if has context
        """
        if meta is None:
            return False
        return meta.has_additional_ctx()

    def current_has_context(self) -> bool:
        """
        Check if current context has additional context from attachments

        :return: True if has context
        """
        meta = self.window.core.ctx.get_current_meta()
        if meta is None:
            return False
        return self.has_context(meta)

    def get_mode(self) -> str:
        """
        Get additional context append mode

        :return: Additional context append mode
        """
        return self.mode

    def get_context(
            self,
            ctx: CtxItem,
            history: List[CtxItem]
    ) -> str:
        """
        Get additional context for attachment

        :param ctx: CtxItem instance
        :param history Context items (history)
        :return: Additional context
        """
        if self.mode != self.MODE_DISABLED:
            if self.is_verbose():
                print("\nPreparing additional context...\nContext Mode: {}".format(self.mode))

        self.window.core.attachments.context.reset()  # reset used files and urls

        # get additional context from attachments
        content = self.window.core.attachments.context.get_context(self.mode, ctx, history)

        # append used files and urls to context
        files = self.window.core.attachments.context.get_used_files()
        urls = self.window.core.attachments.context.get_used_urls()
        if files:
            ctx.files = files
        if urls:
            ctx.urls = urls

        if content:
            if self.is_verbose():
                print("\n--- Using additional context ---\n\n{}".format(content))
            return "====================================\nADDITIONAL CONTEXT FROM ATTACHMENT(s): {}".format(content)
        return ""

    def update(self):
        """Update attachments list"""
        mode = self.window.core.config.get("mode")
        if mode == MODE_ASSISTANT:
            self.hide_uploaded()  # hide uploaded attachments in Assistant mode
            return
        meta = self.window.core.ctx.get_current_meta()
        self.update_list(meta)

    def update_list(self, meta: CtxMeta):
        """
        Update list of attachments

        :param meta: CtxMeta instance
        """
        # update list of attachments
        if meta is None or not meta.has_additional_ctx():
            items = []
        else:
            items = self.window.core.attachments.context.get_all(meta)
        self.window.ui.chat.input.attachments_ctx.update(items)
        self.update_tab(meta)

    def update_tab(self, meta: CtxMeta):
        """
        Update tab label

        :param meta: CtxMeta instance
        """
        if meta is None or not meta.has_additional_ctx():
            num_files = 0
        else:
            num_files = self.window.core.attachments.context.count(meta)
        suffix = ''
        if num_files > 0:
            suffix = f' ({num_files})'
        self.window.ui.tabs['input'].setTabText(
            3,
            trans('attachments_uploaded.tab') + suffix,
        )
        """
        if num_files > 0:
           self.show_uploaded()
        else:
           self.hide_uploaded()
        """

    def is_verbose(self) -> bool:
        """
        Check if verbose mode is enabled

        :return: True if verbose mode is enabled
        """
        return self.window.core.config.get("ctx.attachment.verbose", False)

    def show_uploaded(self):
        """Show uploaded attachments"""
        self.window.ui.tabs['input'].setTabVisible(self.uploaded_tab_idx, True)

    def hide_uploaded(self):
        """Hide uploaded attachments"""
        self.window.ui.tabs['input'].setTabVisible(self.uploaded_tab_idx, False)

    def delete_by_idx(
            self,
            idx: int,
            force: bool = False,
            remove_local: bool = True
    ):
        """
        Delete attachment by index

        :param idx: Index on list
        :param force: Force delete
        :param remove_local: Remove local copies
        """
        if not force:
            self.window.ui.dialogs.confirm(
                type='attachments.ctx.delete',
                id=idx,
                msg=trans('attachments.delete.confirm'),
            )
            return
        meta = self.window.core.ctx.get_current_meta()
        if meta is None or not meta.has_additional_ctx():
            return
        items = self.window.core.attachments.context.get_all(meta)
        if idx < len(items):
            item = items[idx]
            self.window.core.attachments.context.delete(meta, item, delete_files=remove_local)
            self.update_list(meta)
            self.window.controller.ctx.update()

    def clear(
            self,
            force: bool = False,
            remove_local: bool = False,
            auto: bool = False
    ):
        """
        Clear attachments list

        :param force: force clear
        :param remove_local: remove local copies
        :param auto: auto clear
        """
        if not force:
            self.window.ui.dialogs.confirm(
                type='attachments.ctx.clear',
                id=-1,
                msg=trans('attachments.clear.confirm'),
            )
            return

        meta = self.window.core.ctx.get_current_meta()
        if meta is None or not meta.has_additional_ctx():
            return
        self.window.core.attachments.context.clear(meta, delete_files=remove_local)
        self.update_list(meta)
        self.window.controller.ctx.update()

    def select(self, idx: int):
        """
        Select uploaded file

        :param idx: index of file
        """
        pass

    def open_by_idx(self, idx: int):
        """
        Open attachment by index

        :param idx: Index on list
        """
        meta = self.window.core.ctx.get_current_meta()
        if meta is None or not meta.has_additional_ctx():
            return
        items = self.window.core.attachments.context.get_all(meta)
        if idx < len(items):
            item = items[idx]
            path = item["path"]
            if "real_path" in item:
                path = item["real_path"]
            if os.path.exists(path) and os.path.isfile(path):
                print("Opening attachment: {}".format(path))
                self.window.controller.files.open(path)

    def open_dir_src_by_idx(self, idx: int):
        """
        Open source directory by index

        :param idx: Index on list
        """
        meta = self.window.core.ctx.get_current_meta()
        if meta is None or not meta.has_additional_ctx():
            return
        items = self.window.core.attachments.context.get_all(meta)
        if idx < len(items):
            item = items[idx]
            path = item["path"]
            if "real_path" in item:
                path = item["real_path"]
            dir = os.path.dirname(path)
            if os.path.exists(dir) and os.path.isdir(dir):
                print("Opening source directory: {}".format(dir))
                self.window.controller.files.open(dir)

    def open_dir_dest_by_idx(self, idx: int):
        """
        Open destination directory by index

        :param idx: Index on list
        """
        meta = self.window.core.ctx.get_current_meta()
        if meta is None or not meta.has_additional_ctx():
            return
        items = self.window.core.attachments.context.get_all(meta)
        if idx < len(items):
            item = items[idx]
            root_dir = self.window.core.attachments.context.get_dir(meta)
            dir = os.path.join(root_dir, item["uuid"])
            if os.path.exists(dir) and os.path.isdir(dir):
                self.window.controller.files.open(dir)
                print("Opening destination directory: {}".format(dir))

    def has_file_by_idx(self, idx: int) -> bool:
        """
        Check if has file by index

        :param idx: Index on list
        :return: True if has file
        """
        meta = self.window.core.ctx.get_current_meta()
        if meta is None or not meta.has_additional_ctx():
            return False
        items = self.window.core.attachments.context.get_all(meta)
        if idx < len(items):
            item = items[idx]
            path = item["path"]
            if "real_path" in item:
                path = item["real_path"]
            return os.path.exists(path) and os.path.isfile(path)
        return False

    def has_src_by_idx(self, idx: int) -> bool:
        """
        Check if has source directory by index

        :param idx: Index on list
        :return: True if has source directory
        """
        meta = self.window.core.ctx.get_current_meta()
        if meta is None or not meta.has_additional_ctx():
            return False
        items = self.window.core.attachments.context.get_all(meta)
        if idx < len(items):
            item = items[idx]
            path = item["path"]
            if "real_path" in item:
                path = item["real_path"]
            dir = os.path.dirname(path)
            return os.path.exists(dir) and os.path.isdir(dir)
        return False

    def has_dest_by_idx(self, idx: int) -> bool:
        """
        Check if has destination directory by index

        :param idx: Index on list
        :return: True if has destination directory
        """
        meta = self.window.core.ctx.get_current_meta()
        if meta is None or not meta.has_additional_ctx():
            return False
        items = self.window.core.attachments.context.get_all(meta)
        if idx < len(items):
            item = items[idx]
            root_dir = self.window.core.attachments.context.get_dir(meta)
            dir = os.path.join(root_dir, item["uuid"])
            return os.path.exists(dir) and os.path.isdir(dir)
        return False

    def get_current_tokens(self) -> int:
        """
        Get current tokens

        :return: Current attachments tokens
        """
        if self.mode != self.MODE_FULL_CONTEXT:
            return 0
        meta = self.window.core.ctx.get_current_meta()
        if meta is None:
            return 0
        if not meta.has_additional_ctx():
            return 0
        tokens = 0
        for item in meta.get_additional_ctx():
            if "tokens" in item:
                try:
                    tokens += int(item["tokens"])
                except Exception as e:
                    pass
        return tokens

    @Slot(object)
    def handle_upload_error(self, error: Exception):
        """
        Handle upload error

        :param error: Exception
        """
        self.window.dispatch(KernelEvent(KernelEvent.STATE_ERROR, {
            "id": "chat",
            "msg": "Error reading attachments: {}".format(str(error))
        }))

    @Slot(str)
    def handle_upload_success(self, text: str):
        """
        Handle upload success

        :param text: Input prompt text
        """
        context = BridgeContext()
        context.prompt = text
        event = KernelEvent(KernelEvent.INPUT_USER, {
            'context': context,
            'extra': {},
        })
        self.window.dispatch(event)

    def switch_mode(self, mode: str):
        """
        Switch context mode

        :param mode: context mode
        """
        self.mode = mode
        self.window.core.config.set("ctx.attachment.mode", mode)
        self.window.core.config.save()
        self.window.controller.ui.update_tokens()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.18 21:00:00                  #
# ================================================== #

import os.path
import subprocess

from datetime import datetime
from PySide6.QtCore import Slot

from pygpt_net.plugin.base.worker import BaseWorker, BaseSignals


class WorkerSignals(BaseSignals):
    pass  # add custom signals here


class Worker(BaseWorker):
    def __init__(self, *args, **kwargs):
        super(Worker, self).__init__()
        self.signals = BaseSignals()
        self.args = args
        self.kwargs = kwargs
        self.plugin = None
        self.cmds = None
        self.ctx = None

    @Slot()
    def run(self):
        responses = []
        msg = None
        for item in self.cmds:
            if self.is_stopped():
                break
            for my_cmd in self.plugin.get_option_value("cmds"):
                if self.is_stopped():
                    break
                if my_cmd["name"] == item["cmd"]:
                    try:
                        response = self.handle_cmd(my_cmd, item)
                        if response is False:
                            msg = "Command is empty"
                            continue
                        responses.append(response)

                    except Exception as e:
                        responses.append(
                            self.make_response(
                                item,
                                self.throw_error(e)
                            )
                        )

        # send response
        if len(responses) > 0:
            self.reply_more(responses)

        # update status
        if msg is not None:
            self.status(msg)

    def handle_cmd(self, command: dict, item: dict) -> dict or bool:
        """
        Handle custom command

        :param command: command configuration
        :param item: command item
        :return: response item
        """
        # prepare cmd
        cmd = command["cmd"]

        # check if cmd is not empty
        if cmd is None or cmd == "":
            return False
        
        try:
            cmd = cmd.format(
                _file=os.path.dirname(os.path.realpath(__file__)),
                _home=self.plugin.window.core.config.path,
                _date=datetime.now().strftime("%Y-%m-%d"),
                _time=datetime.now().strftime("%H:%M:%S"),
                _datetime=datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            )
        except Exception as e:
            pass

        # append params to placeholders
        if 'params' in command and command["params"].strip() != "":
            # append params to cmd placeholders
            params_list = self.plugin.extract_params(
                command["params"],
            )
            for param in params_list:
                if param in item["params"]:
                    cmd = cmd.replace(
                        "{" + param + "}",
                        str(item["params"][param]),
                    )

        # execute
        msg = "Running custom cmd: {}".format(cmd)
        self.log(msg)
        process = subprocess.Popen(
            cmd,
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        stdout, stderr = process.communicate()
        result = None
        if stdout:
            result = stdout.decode("utf-8")
            self.log("STDOUT: {}".format(result))
        if stderr:
            result = stderr.decode("utf-8")
            self.log("STDERR: {}".format(result))
        if result is None:
            result = "No result (STDOUT/STDERR empty)"
            self.log(result)

        return self.make_response(item, result)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 08:00:00                  #
# ================================================== #

import copy
import json
import os

from typing import Optional


class Editor:
    def __init__(self, window=None):
        """
        File editor core

        :param window: Window instance
        """
        self.window = window

    def restore(self, dialog_id: str):
        """
        Load defaults from file

        :param dialog_id: dialog id
        """
        file = self.window.ui.dialog[dialog_id].file
        self.load(dialog_id, file)
        self.window.update_status("Reloaded file: {}".format(file))

    def clear(self, dialog_id: str):
        """
        Clear editor

        :param dialog_id: dialog id
        """
        self.window.ui.editor[dialog_id].clear()
        self.window.ui.dialog[dialog_id].file = None
        self.window.ui.dialog[dialog_id].base_content = ""
        self.window.ui.dialog[dialog_id].reset_file_title()
        self.window.ui.editor[dialog_id].on_update()
        self.window.update_status("")

    def destroy(self, dialog_id: str):
        """
        On destroy editor

        :param dialog_id: dialog id
        """
        self.window.ui.editor[dialog_id].on_destroy()

    def is_changed(self, dialog_id: str) -> bool:
        """
        Check if content was changed

        :param dialog_id: dialog id
        :return: True if changed
        """
        return self.window.ui.dialog[dialog_id].is_changed()

    def load(
            self,
            dialog_id: str,
            file: Optional[str] = None
    ):
        """
        Load file to editor

        :param dialog_id: dialog id
        :param file: file name
        """
        # load file
        self.window.ui.paths[dialog_id].setText(file)
        self.window.ui.dialog[dialog_id].file = file
        if not os.path.exists(file):
            self.window.update_status("File not found: {}".format(file))
            return
        try:
            with open(file, 'r', encoding="utf-8") as f:
                txt = f.read()
                self.window.ui.editor[dialog_id].setPlainText(txt)
                self.window.ui.dialog[dialog_id].base_content = copy.deepcopy(txt)
                self.window.ui.dialog[dialog_id].update_file_title()
                self.window.ui.editor[dialog_id].on_update()
        except Exception as e:
            self.window.core.debug.log(e)
            self.window.update_status("Error loading file: {}".format(e))

    def save(
            self,
            dialog_id: str,
            path: Optional[str] = None
    ):
        """
        Save file to disk

        :param dialog_id: dialog id
        :param path: file path
        """
        file = self.window.ui.dialog[dialog_id].file
        if path is None:
            path = file
        else:
            self.window.ui.dialog[dialog_id].file = path  # update file path

        data = self.window.ui.editor[dialog_id].toPlainText()

        # check if this is a valid JSON
        if path.endswith('.json'):
            try:
                json.loads(data)
            except Exception as e:
                self.window.update_status("WARNING: This is not a valid JSON: {}".format(e))
                self.window.ui.dialogs.alert("WARNING: This is not a valid JSON: {}".format(e))

        # save changes
        try:
            with open(path, 'w', encoding="utf-8") as f:
                f.write(data)
            self.window.ui.dialog[dialog_id].base_content = copy.deepcopy(data)
            self.window.ui.dialog[dialog_id].update_file_title(force=True)
            self.window.update_status("Saved file: {}".format(os.path.basename(path)))
            self.window.controller.files.update_explorer()
        except Exception as e:
            self.window.core.debug.log(e)
            self.window.update_status("Error saving file: {}".format(path))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.01.16 01:00:00                  #
# ================================================== #

from typing import Optional, List, Dict

from llama_index.llms.gemini import Gemini
from llama_index.core.llms.llm import BaseLLM as LlamaBaseLLM
from llama_index.core.base.embeddings.base import BaseEmbedding
from llama_index.embeddings.gemini import GeminiEmbedding

from pygpt_net.core.types import (
    MODE_LLAMA_INDEX,
)
from pygpt_net.provider.llms.base import BaseLLM
from pygpt_net.item.model import ModelItem


class GoogleLLM(BaseLLM):
    def __init__(self, *args, **kwargs):
        super(GoogleLLM, self).__init__(*args, **kwargs)
        """
        Required ENV variables:
            - GOOGLE_API_KEY - API key for Google API
        Required args:
            - model: model name, e.g. gemini-1,5-pro
            - api_key: API key for Google API
        """
        self.id = "google"
        self.type = [MODE_LLAMA_INDEX, "embeddings"]

    def llama(
            self,
            window,
            model: ModelItem,
            stream: bool = False
    ) -> LlamaBaseLLM:
        """
        Return LLM provider instance for llama

        :param window: window instance
        :param model: model instance
        :param stream: stream mode
        :return: LLM provider instance
        """
        args = self.parse_args(model.llama_index)
        return Gemini(**args)

    def get_embeddings_model(
            self,
            window,
            config: Optional[List[Dict]] = None
    ) -> BaseEmbedding:
        """
        Return provider instance for embeddings

        :param window: window instance
        :param config: config keyword arguments list
        :return: Embedding provider instance
        """
        args = {}
        if config is not None:
            args = self.parse_args({
                "args": config,
            })
        return GeminiEmbedding(**args)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.01.19 02:00:00                  #
# ================================================== #

from PySide6.QtCore import QRect, QDate
from PySide6.QtGui import QColor, QBrush, QFont, Qt, QAction, QContextMenuEvent, QIcon, QPixmap, QPen
from PySide6.QtWidgets import QCalendarWidget, QMenu

from pygpt_net.core.tabs.tab import Tab
from pygpt_net.utils import trans
import pygpt_net.icons_rc


class CalendarSelect(QCalendarWidget):
    def __init__(self, window=None):
        """
        Calendar select widget

        :param window: main window
        """
        super(CalendarSelect, self).__init__(window)
        self.window = window
        self.currentYear = QDate.currentDate().year()
        self.currentMonth = QDate.currentDate().month()
        self.currentDay = QDate.currentDate().day()
        self.font_size = 8
        self.counters = {
            'ctx': {},  # num of ctx in date
            'notes': {},  # num of notes in date
        }
        self.labels = {}
        self.setGridVisible(True)
        self.setVerticalHeaderFormat(QCalendarWidget.NoVerticalHeader)  # disable num of weeks display
        self.currentPageChanged.connect(self.page_changed)
        self.clicked[QDate].connect(self.on_day_clicked)

        # context menu
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self.open_context_menu)
        self.setProperty('class', 'calendar')
        self.tab = None
        self.installEventFilter(self)

    def set_tab(self, tab: Tab):
        """
        Set tab

        :param tab: Tab
        """
        self.tab = tab

    def eventFilter(self, source, event):
        """
        Focus event filter

        :param source: source
        :param event: event
        """
        if event.type() == event.Type.FocusIn:
            if self.tab is not None:
                col_idx = self.tab.column_idx
                self.window.controller.ui.tabs.on_column_focus(col_idx)
        return super().eventFilter(source, event)

    def page_changed(self, year, month):
        """
        On page changed

        :param year: Year
        :param month: Month
        """
        self.currentYear = year
        self.currentMonth = month
        self.window.controller.calendar.on_page_changed(year, month)

    def paintCell(self, painter, rect, date: QDate):
        """
        On painting cell

        :param painter: Painter
        :param rect: Rectangle
        :param date: Date
        """

        super().paintCell(painter, rect, date)

        # current date
        if date == QDate.currentDate():
            painter.save()
            pen = QPen(QColor(0, 0, 0))
            pen.setWidth(2)
            painter.setPen(pen)
            painter.drawRect(rect)
            painter.restore()

        # ctx counter
        if date in self.counters['ctx']:
            padding = 2
            task_rect = QRect(
                rect.right() - padding - 20,
                rect.top() + padding,
                20,
                20,
            )
            painter.save()
            painter.setBrush(QBrush(QColor(40, 40, 40)))
            painter.drawRect(task_rect)
            painter.setPen(QColor(255, 255, 255))
            painter.setFont(QFont('Lato', self.font_size))
            painter.drawText(
                task_rect,
                Qt.AlignCenter,
                str(self.counters['ctx'][date]),
            )
            painter.restore()

        # notes counter
        if date in self.counters['notes']:
            day_notes = self.counters['notes'][date]
            for status, count in day_notes.items():
                padding = 2
                task_rect = QRect(
                    rect.left() + padding,
                    rect.bottom() - padding - 20,
                    20,
                    20,
                )
                painter.save()
                bg_color, font_color = self.get_color_for_status(status)
                painter.setBrush(QBrush(bg_color))
                painter.drawRect(task_rect)
                painter.setPen(font_color)
                painter.setFont(QFont('Lato', self.font_size))
                painter.drawText(
                    task_rect,
                    Qt.AlignCenter,
                    "!",
                )  # str(count)
                painter.restore()

        if date in self.labels:
            # draw little square with color if label exists in date
            prev_left = rect.left()
            for label_id in self.labels[date]:
                colors = self.window.controller.ui.get_colors()
                color = colors[label_id]['color']
                painter.save()
                pen = QPen(QColor(0, 0, 0))
                pen.setWidth(1)
                painter.setPen(pen)
                painter.setBrush(QBrush(color))
                painter.drawRect(
                    prev_left + 2,
                    rect.top() + 2,
                    5,
                    5,
                )
                painter.restore()
                prev_left += 7

    def get_color_for_status(self, status: int) -> (QColor, QColor):
        """
        Get color for status

        :param status: status
        :return: color, font color
        """
        colors = self.window.controller.ui.get_colors()
        if status in colors:
            return colors[status]['color'], colors[status]['font']
        else:
            return QColor(100, 100, 100), QColor(255, 255, 255)

    def on_day_clicked(self, date: QDate):
        """
        On day clicked

        :param date: Date
        """
        year = date.year()
        month = date.month()
        day = date.day()
        self.currentYear = year
        self.currentMonth = month
        self.currentDay = day
        self.window.controller.calendar.on_day_select(year, month, day)

        # check if date has ctx TODO: think about better solution
        # if date in self.counters['ctx']:
        self.window.controller.calendar.on_ctx_select(year, month, day)

        if self.tab is not None:
            col_idx = self.tab.column_idx
            self.window.controller.ui.tabs.on_column_focus(col_idx)

    def add_ctx(self, date: QDate, num: int):
        """
        Add ctx counter to counter list

        :param date: date
        :param num: number of ctx
        """
        self.counters['ctx'][date] = str(num)
        self.updateCell(date)

    def update_ctx(self, counters: dict, labels: dict):
        """
        Update ctx counters

        :param counters: counters dict
        :param labels: labels dict
        """
        self.counters['ctx'] = {
            QDate.fromString(date_str, 'yyyy-MM-dd'): count for date_str, count in counters.items()
        }
        self.labels = {
            QDate.fromString(date_str, 'yyyy-MM-dd'): labels for date_str, labels in labels.items()
        }
        self.updateCells()

    def update_notes(self, counters: dict):
        """
        Update notes counters

        :param counters: counters dict
        """
        self.counters['notes'] = {
            QDate.fromString(date_str, 'yyyy-MM-dd'): count for date_str, count in counters.items()
        }
        self.updateCells()

    def open_context_menu(self, position):
        """
        Open context menu

        :param position: position
        """
        colors = self.window.controller.ui.get_colors()
        selected_date = self.selectedDate()
        context_menu = QMenu(self)
        action_text = trans('calendar.day.search') + ': ' + selected_date.toString()
        action = QAction(action_text, self)
        action.setIcon(QIcon(":/icons/history.svg"))
        action.triggered.connect(lambda: self.execute_action(selected_date))
        context_menu.addAction(action)
        
        # set label menu
        set_label_menu = context_menu.addMenu(trans('calendar.day.label'))
        for status_id, status_info in colors.items():
            name = trans('calendar.day.' + status_info['label'])
            if status_id == 0:
                name = '-'
            color = status_info['color']
            pixmap = QPixmap(16, 16)
            pixmap.fill(color)
            icon = QIcon(pixmap)
            status_action = QAction(icon, name, self)
            status_action.triggered.connect(
                lambda checked=False, s_id=status_id: self.set_label_for_day(selected_date, s_id)
            )
            set_label_menu.addAction(status_action)

        context_menu.exec(self.mapToGlobal(position))

    def execute_action(self, date):
        """
        On select date from context menu

        :param date: QDate
        """
        year = date.year()
        month = date.month()
        day = date.day()
        self.window.controller.calendar.on_ctx_select(
            year,
            month,
            day,
        )

    def contextMenuEvent(self, event: QContextMenuEvent):
        """
        On context menu event

        :param event: context menu event
        """
        self.open_context_menu(event.pos())

    def set_label_for_day(self, date: QDate, status_id: int):
        """
        Set label for day

        :param date: date
        :param status_id: status id
        """
        self.window.controller.calendar.note.update_status(
            status_id,
            date.year(),
            date.month(),
            date.day(),
        )




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.03.02 19:00:00                  #
# ================================================== #

from pygpt_net.core.types import (
    MODE_AGENT,
    MODE_AGENT_LLAMA,
    MODE_ASSISTANT,
    MODE_AUDIO,
    MODE_CHAT,
    MODE_LANGCHAIN,
    MODE_LLAMA_INDEX,
    MODE_VISION,
    MODE_RESEARCH,
)
from pygpt_net.plugin.base.plugin import BasePlugin
from pygpt_net.core.bridge.context import BridgeContext
from pygpt_net.item.model import ModelItem
from pygpt_net.core.events import Event
from pygpt_net.item.ctx import CtxItem

from .config import Config


class Plugin(BasePlugin):
    def __init__(self, *args, **kwargs):
        super(Plugin, self).__init__(*args, **kwargs)
        self.id = "openai_dalle"
        self.name = "DALL-E 3: Image generation"
        self.description = "Integrates DALL-E 3 image generation with any chat"
        self.prefix = "DALL-E"
        self.type = [
            "cmd.inline",
        ]
        self.allowed_modes = [
            MODE_CHAT,
            MODE_LANGCHAIN,
            MODE_VISION,
            MODE_LLAMA_INDEX,
            MODE_ASSISTANT,
            MODE_AGENT,
            MODE_AUDIO,
            MODE_RESEARCH,
        ]
        self.allowed_cmds = [
            "image",
        ]
        self.order = 100
        self.use_locale = True
        self.config = Config(self)
        self.init_options()

    def init_options(self):
        """Initialize options"""
        self.config.from_defaults(self)

    def handle(self, event: Event, *args, **kwargs):
        """
        Handle dispatched event

        :param event: event object
        :param args: args
        :param kwargs: kwargs
        """
        name = event.name
        data = event.data
        ctx = event.ctx

        if name == Event.SYSTEM_PROMPT:
            if self.is_native_cmd():  # only if native commands are disabled, otherwise use commands only
                return

            mode = ""
            if "mode" in data:
                mode = data["mode"]
            if mode not in self.allowed_modes:
                return
            data['value'] = self.on_system_prompt(data['value'])

        elif name in [
            Event.CMD_SYNTAX_INLINE,  # inline is allowed
            Event.CMD_SYNTAX,
        ]:
            if not self.is_native_cmd() and "force" not in data:  # only if native commands are enabled, otherwise use prompt only
                return

            self.cmd_syntax(data)

        elif name in [
            Event.CMD_INLINE,
            Event.CMD_EXECUTE,
        ]:
            self.cmd(
                ctx,
                data['commands'],
            )

    def cmd_syntax(self, data: dict):
        """
        Event: CMD_SYNTAX or CMD_SYNTAX_INLINE

        :param data: event data dict
        """
        for option in self.allowed_cmds:
            if not self.has_cmd(option):
                continue
            data['cmd'].append(self.get_cmd(option))  # append command

    def on_system_prompt(self, prompt: str) -> str:
        """
        Event: SYSTEM_PROMPT

        :param prompt: prompt
        :return: updated prompt
        """
        prompt += "\n" + self.get_option_value("prompt")
        return prompt

    def cmd(self, ctx: CtxItem, cmds: list):
        """
        Events: CMD_INLINE, CMD_EXECUTE

        :param ctx: CtxItem
        :param cmds: commands dict
        """
        is_cmd = False
        my_commands = []
        for item in cmds:
            if item["cmd"] in self.allowed_cmds:
                my_commands.append(item)
                is_cmd = True

        if not is_cmd:
            return

        for item in my_commands:
            try:
                if item["cmd"] == "image":
                    query = item["params"]["query"]
                    # if internal call (ctx.internal = True), then re-send OK response
                    # if not internal call, then append image to chat only
                    model = ModelItem()
                    model.id = "dall-e-3"
                    bridge_context = BridgeContext(
                        ctx=ctx,
                        mode="image",  # fake mode, not-img mode
                        model=model,  # model instance
                        prompt=query,
                    )
                    extra = {
                        "num": 1,  # force 1 image if dall-e-3 model is used
                        "inline": True, # force inline mode
                    }
                    sync = False
                    if self.window.core.config.get("mode") == MODE_AGENT_LLAMA:
                        sync = True
                    self.window.core.gpt.image.generate(bridge_context, extra, sync)  # force inline mode, async call
            except Exception as e:
                self.log("Error: " + str(e))
                return



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.05 23:00:00                  #
# ================================================== #

import re
import html


class Helpers:

    def __init__(self, window=None):
        """
        Helpers

        :param window: Window instance
        """
        self.window = window

    def replace_code_tags(self, text: str) -> str:
        """
        Replace cmd code tags

        :param text:
        :return: replaced text
        """
        pattern = r"~###~(.*?)~###~"
        replacement = r'<p class="cmd">\1</p>'
        return re.sub(pattern, replacement, text)

    def pre_format_text(self, text: str) -> str:
        """
        Pre-format text

        :param text: text to format
        :return: formatted text
        """
        text = text.strip()
        text = text.replace("#~###~", "~###~")  # fix for #~###~ in text (previous versions)
        text = text.replace("# ~###~", "~###~")  # fix for # ~###~ in text (previous versions)

        # replace cmd tags
        text = self.replace_code_tags(text)

        # replace %workdir% with current workdir
        local_prefix = self.window.core.filesystem.get_workdir_prefix()
        safe_local_prefix = local_prefix.replace('\\', '\\\\').replace('\\.', '\\\\.')  # windows fix
        replacement = f'({safe_local_prefix}\\1)'
        try:
            text = re.sub(r'\(%workdir%([^)]+)\)', replacement, text)
        except Exception as e:
            pass
        return text

    def post_format_text(self, text: str) -> str:
        """
        Post-format text

        :param text: text to format
        :return: formatted text
        """
        return text.strip()

    def format_user_text(self, text: str) -> str:
        """
        Post-format user text

        :param text: text to format
        :return: formatted text
        """
        text = html.escape(text).replace("\n", "<br>")

        # append cmd tags if response from command detected
        if text.strip().startswith("&gt; [") and text.strip().endswith("]"):
            text = '<div class="cmd">&gt; {}</div><br/>'.format(text)
        return text

    def format_chunk(self, text: str) -> str:
        """
        Format chunk

        :param text: text to format
        :return: formatted text
        """
        return text



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 08:00:00                  #
# ================================================== #

import os
from typing import Any, Optional

from pygpt_net.core.events import RenderEvent

from .common import Common
from .markdown import Markdown
from .menu import Menu
from .nodes import Nodes

class Theme:
    def __init__(self, window=None):
        """
        Theme controller

        :param window: Window instance
        """
        self.window = window
        self.common = Common(window)
        self.markdown = Markdown(window)
        self.menu = Menu(window)
        self.nodes = Nodes(window)

    def setup(self):
        """Setup theme"""
        # load markdown CSS
        self.markdown.load()

        # setup menus
        self.menu.setup_list()
        self.menu.setup_density()
        self.menu.setup_syntax()

        # show or hide tooltips
        self.common.toggle_tooltips()

        # apply current theme to nodes
        self.reload(force=False)

    def toggle(
            self,
            name: str,
            force: bool = True
    ):
        """
        Toggle theme by name

        :param name: theme name
        :param force: force theme change (manual trigger)
        """
        if force:
            self.window.controller.ui.store_state()  # store state before theme change

        self.window.core.config.set('theme', name)
        self.window.core.config.save()
        self.nodes.apply_all()
        self.apply(name + '.xml', self.common.get_extra_css(name))  # style.css = additional custom stylesheet

        # apply markdown CSS
        self.markdown.update(force=False)

        # update themes menu
        self.menu.update_list()
        self.menu.update_syntax()

        if force:
            self.window.controller.ui.restore_state()  # restore state after theme change

    def toggle_style(self, name: str):
        """
        Toggle theme style (web)

        :param name: web style name
        """
        self.window.core.config.set('theme.style', name)
        self.window.core.config.save()
        event = RenderEvent(RenderEvent.ON_THEME_CHANGE)
        self.window.dispatch(event)
        self.reload()

    def toggle_option(
            self,
            name: str,
            value: Any = None
    ):
        """
        Toggle theme menu option

        :param name: option name
        :param value: option value
        """
        if name == 'layout.tooltips':
            if self.window.core.config.get(name):
                state = False
            else:
                state = True
            self.window.core.config.set(name, state)
            self.window.controller.config.checkbox.apply('config', 'layout.tooltips', {'value': state})
            self.common.toggle_tooltips()
        elif name == 'layout.density':
            value = int(value)
            self.window.core.config.set(name, value)
            self.window.controller.config.slider.apply('config', 'layout.density', {'value': value})
            self.reload()
            self.menu.update_density()
        elif name == 'render.blocks':
            if self.window.core.config.get(name):
                state = False
            else:
                state = True
            self.window.core.config.set(name, state)
            event = RenderEvent(RenderEvent.ON_THEME_CHANGE)
            self.window.dispatch(event)
            self.reload()
        self.window.core.config.save()
        self.nodes.apply_all()

    def toggle_syntax(
            self,
            name: str,
            update_menu: bool = False
    ):
        """
        Toggle syntax highlight

        :param name: syntax style name
        :param update_menu: update menu
        """
        self.window.core.config.set("render.code_syntax", name)
        self.window.core.config.save()
        event = RenderEvent(RenderEvent.ON_THEME_CHANGE)
        self.window.dispatch(event)
        if update_menu:
            self.menu.update_syntax()

    def update_style(self):
        """Update style"""
        current = self.window.core.config.get('theme.style')
        self.toggle_style(current)

    def update_theme(self, force: bool = True):
        """
        Update theme

        :param force: force theme change (manual trigger)
        """
        current = self.window.core.config.get('theme')
        self.toggle(current, force=force)

    def update_syntax(self):
        """Update syntax menu"""
        current = self.window.core.config.get('render.code_syntax')
        self.toggle_syntax(current, update_menu=True)

    def reload(self, force: bool = True):
        """
        Reload current theme

        :param force: force theme change (manual trigger)
        """
        self.update_theme(force=force)

    def apply(
            self,
            theme: str = 'dark_teal.xml',
            custom: Optional[str] = None
    ):
        """
        Update material theme and apply custom CSS

        :param theme: material theme filename (e.g. dark_teal.xml)
        :param custom: additional custom stylesheet filename (e.g. style.css)
        """
        inverse = False
        if theme.startswith('light'):
            inverse = True
        extra = {
            'density_scale': self.window.core.config.get('layout.density'),
            'pyside6': True,
        }
        self.window.apply_stylesheet(self.window, theme, invert_secondary=inverse, extra=extra)

        # append custom stylesheet
        if custom is not None:
            stylesheet = self.window.styleSheet()  # get current stylesheet
            paths = []
            paths.append(os.path.join(self.window.core.config.get_app_path(), 'data', 'css', custom))
            paths.append(os.path.join(self.window.core.config.get_user_path(), 'css', custom))
            content = ''
            for path in paths:
                if os.path.exists(path):
                    with open(path) as file:
                        content += file.read()

            # Windows checkbox button + radio button fix (if no SVG support)
            # when missing DLLs in PySide6, VC++ redistributable required
            if self.window.core.platforms.is_windows() and not self.window.core.config.is_compiled():
                content += self.common.get_windows_fix()
            try:
                self.window.setStyleSheet(stylesheet + content.format(**os.environ))  # apply stylesheet
            except KeyError as e:
                pass  # ignore missing env variables

    def style(self, element: str) -> str:
        """
        Return CSS style for element (alias)

        :param element: type of element
        :return: CSS style for element
        """
        return self.common.get_style(element)

    def reload_all(self):
        """Reload all"""
        self.setup()
        self.update_style()
        self.update_syntax()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 08:00:00                  #
# ================================================== #

from typing import Optional
import webbrowser

from pygpt_net.item.assistant import AssistantItem

from .batch import Batch
from .editor import Editor
from .files import Files
from .store import VectorStore
from .threads import Threads

from pygpt_net.core.text.utils import has_unclosed_code_tag
from pygpt_net.utils import trans
from pygpt_net.item.ctx import CtxItem
from pygpt_net.core.events import RenderEvent


class Assistant:
    def __init__(self, window=None):
        """
        Assistant controller

        :param window: Window instance
        """
        self.window = window
        self.batch = Batch(window)
        self.editor = Editor(window)
        self.files = Files(window)
        self.threads = Threads(window)
        self.store = VectorStore(window)

    def setup(self):
        """Setup assistants"""
        self.window.core.assistants.load()
        self.editor.setup()
        self.update()

    def update(self, update_list: bool = True):
        """
        Update assistants list

        :param update_list: True if update list
        """
        if update_list:
            self.update_list()
        self.files.update_list()
        self.select_current()

    def update_list(self):
        """Update assistants list"""
        items = self.window.core.assistants.get_all()
        self.window.ui.toolbox.assistants.update(items)

    def run_stop(self):
        """Stop assistant run"""
        ctx = self.window.core.ctx.get_last_item()
        if ctx is not None:
            if ctx.run_id is not None and ctx.thread is not None:
                ctx.stopped = True  # mark as canceled
                self.threads.log("FORCE STOP: {}".format(ctx.run_id))
                if has_unclosed_code_tag(ctx.output):
                    ctx.output += "\n```"  # fix for code block without closing ```
                self.window.core.ctx.update_item(ctx)  # save current output

                # get status
                try:
                    status = self.window.core.gpt.assistants.run_stop(ctx)
                    if status == "cancelling" or status == "cancelled":
                        print("Run has been canceled.")
                        self.threads.log("Run status: {}".format(status))
                except Exception as e:
                    print("Run stop failed: ", e)

                # render final output
                event = RenderEvent(RenderEvent.STREAM_END, {
                    "meta": ctx.meta,
                    "ctx": ctx,
                })
                self.window.dispatch(event)
                self.window.controller.assistant.threads.handle_output_message(ctx, stream=True)

    def begin(self, ctx: CtxItem):
        """
        Begin assistants

        :param ctx: CtxItem instance
        """
        ctx.thread = self.prepare()  # create new thread if not exists
        self.window.controller.chat.files.send("assistant", ctx)  # upload attachments

    def prepare(self):
        """
        Prepare assistants

        :return: current thread ID
        """
        # create or get current thread, it is required before conversation start
        if self.window.core.config.get('assistant_thread') is None:
            try:
                thread_id = self.threads.create_thread()
                self.window.update_status(trans('status.sending'))
                self.window.core.config.set('assistant_thread', thread_id)
            except Exception as e:
                self.window.core.debug.log(e)
                self.window.ui.dialogs.alert(e)
        return self.window.core.config.get('assistant_thread')

    def refresh(self):
        """Update assistants"""
        self.select_default()

    def select(self, idx: int):
        """
        Select assistant on the list

        :param idx: idx on the list
        """
        # check if change is not locked
        if self.change_locked():
            return

        # mark assistant as selected
        id = self.window.core.assistants.get_by_idx(idx)
        self.select_by_id(id)

    def select_by_id(self, id: str):
        """
        Select assistant on the list

        :param id: assistant ID
        """
        self.window.core.config.set('assistant', id)

        # update attachments list with list of attachments from assistant
        mode = self.window.core.config.get('mode')
        assistant = self.window.core.assistants.get_by_id(id)
        self.window.controller.attachment.update()
        self.update(False)

        # update model if exists in assistant
        if assistant is not None:
            model = assistant.model
            if model is not None and model != "":
                if model in self.window.core.models.items:
                    self.window.core.config.set('model', model)
                    self.window.core.config.data['current_model'][mode] = model
                    self.window.controller.model.set(mode, model)
                    self.window.controller.model.init_list()
                    self.window.controller.model.select_current()
            self.window.ui.nodes['preset.prompt'].setPlainText(assistant.instructions)
            self.refresh()

        self.window.controller.ctx.update_ctx()  # update current ctx info

    def from_global(self):
        """Update current preset from global prompt"""
        id = self.window.core.config.get('assistant')
        if id is not None and id != "":
            if self.window.core.assistants.has(id):
                assistant = self.window.core.assistants.get_by_id(id)
                assistant.instructions = self.window.core.config.get('prompt')

    def select_current(self):
        """Select assistant by current"""
        assistant_id = self.window.core.config.get('assistant')
        items = self.window.core.assistants.get_all()
        if assistant_id in items:
            idx = list(items.keys()).index(assistant_id)
            current = self.window.ui.models['assistants'].index(idx, 0)
            self.window.ui.nodes['assistants'].setCurrentIndex(current)
            self.window.core.config.set('prompt', items[assistant_id].instructions)
            self.window.ui.nodes['preset.prompt'].setPlainText(items[assistant_id].instructions)

    def select_default(self):
        """Set default assistant"""
        assistant = self.window.core.config.get('assistant')
        if assistant is None or assistant == "":
            mode = self.window.core.config.get('mode')
            if mode == 'assistant':
                self.window.core.config.set(
                    'assistant',
                    self.window.core.assistants.get_default_assistant(),
                )
                self.update()

    def create(self) -> AssistantItem:
        """
        Create assistant in API

        :return: AssistantItem
        """
        assistant = self.window.core.assistants.create()
        self.editor.assign_data(assistant)
        try:
            return self.window.core.gpt.assistants.create(assistant)
        except Exception as e:
            self.window.core.debug.log(e)
            self.window.ui.dialogs.alert(e)

    def update_data(self, assistant: AssistantItem):
        """
        Update assistant

        :param assistant: AssistantItem
        """
        self.editor.assign_data(assistant)
        try:
            return self.window.core.gpt.assistants.update(assistant)
        except Exception as e:
            self.window.core.debug.log(e)
            self.window.ui.dialogs.alert(e)

    def clear(self, force: bool = False):
        """
        Clear assistant data

        :param force: force clear data
        """
        id = self.window.core.config.get('assistant')

        if not force:
            self.window.ui.dialogs.confirm(
                type='assistant_clear',
                id='',
                msg=trans('confirm.assistant.clear'),
            )
            return

        if id is not None and id != "":
            if self.window.core.assistants.has(id):
                assistant = self.window.core.assistants.get_by_id(id)
                assistant.reset()

        self.window.update_status(trans('status.assistant.cleared'))
        self.update()

    def delete(
            self,
            idx: Optional[int] = None,
            force: bool = False
    ):
        """
        Delete assistant

        :param idx: assistant index (row index)
        :param force: force delete without confirmation
        """
        if idx is not None:
            id = self.window.core.assistants.get_by_idx(idx)
            if id is not None and id != "":
                if self.window.core.assistants.has(id):
                    # if exists then show confirmation dialog
                    if not force:
                        self.window.ui.dialogs.confirm(
                            type='assistant.delete',
                            id=idx,
                            msg=trans('confirm.assistant.delete'),
                        )
                        return

                    # clear if this is current assistant
                    if id == self.window.core.config.get('assistant'):
                        self.window.core.config.set('assistant', None)
                        self.window.core.config.set('assistant_thread', None)

                    # delete in API
                    try:
                        self.window.core.gpt.assistants.delete(id)
                    except Exception as e:
                        self.window.ui.dialogs.alert(e)

                    self.window.core.assistants.delete(id)
                    self.update()
                    self.window.update_status(trans('status.assistant.deleted'))

    def goto_online(self):
        """Open Assistants page"""
        webbrowser.open('https://platform.openai.com/assistants')

    def change_locked(self) -> bool:
        """
        Check if assistant change is locked

        :return: True if locked
        """
        return self.window.controller.chat.input.generating

    def reload(self):
        """Reload assistants"""
        self.setup()
        self.store.reset()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.20 21:00:00                  #
# ================================================== #

from pygpt_net.core.bridge.context import BridgeContext
from pygpt_net.plugin.base.plugin import BasePlugin
from pygpt_net.core.events import Event, KernelEvent

from datetime import datetime
from croniter import croniter

from pygpt_net.utils import trans

from .config import Config


class Plugin(BasePlugin):
    def __init__(self, *args, **kwargs):
        super(Plugin, self).__init__(*args, **kwargs)
        self.id = "crontab"
        self.name = "Crontab / Task scheduler"
        self.type = [
            'schedule',
        ]
        self.description = "Plugin provides cron-based job scheduling - " \
                           "you can schedule prompts to be sent at any time using cron-based syntax for task setup."
        self.prefix = "Cron"
        self.order = 100
        self.use_locale = True
        self.timers = []
        self.config = Config(self)
        self.init_options()

    def init_options(self):
        """Initialize options"""
        self.config.from_defaults(self)

    def on_update(self):
        pass

    def on_post_update(self):
        """
        On post update hook
        """
        self.schedule_tasks()

    def count_active(self) -> int:
        """
        Count active tasks

        :return: number of active tasks
        """
        count = 0
        for item in self.get_option_value("crontab"):
            if item["enabled"]:
                count += 1
        return count

    def handle(self, event: Event, *args, **kwargs):
        """
        Handle dispatched event

        :param event: event object
        :param args: event args
        :param kwargs: event kwargs
        """
        name = event.name
        data = event.data

        if name == Event.PLUGIN_SETTINGS_CHANGED:
            self.window.ui.tray.update_schedule_tasks(
                self.count_active()
            )

        elif name == Event.PLUGIN_OPTION_GET:
            if "name" in data and data["name"] == "scheduled_tasks_count":
                data["value"] = self.count_active()  # return number of active tasks

    def job(self, item: dict):
        """
        Execute task

        :param item: task item dict
        """
        if item["prompt"] == "" or item["prompt"] is None:
            self.log("Prompt is empty, skipping task")
            return

        self.log("Executing task: {}: {}".format(datetime.now(), item["prompt"]))

        # show notification
        if self.get_option_value("show_notify"):
            self.window.ui.tray.show_msg(
                trans("notify.cron.title"),
                item["prompt"][0:30] + "...",
            )

        # select preset if defined and exists
        if item["preset"] != "_" and item["preset"] is not None:
            if self.window.core.presets.exists(item["preset"]):
                mode = self.window.core.presets.get_first_mode(item["preset"])
                if mode is not None:
                    self.log("Using preset: {}".format(item["preset"]))
                    self.window.controller.presets.set(mode, item["preset"])
                    self.window.controller.presets.select_current()
                    self.window.controller.presets.select_model()

        # send prompt
        if self.get_option_value("new_ctx"):
            self.window.controller.ctx.new(
                force=True,
            )

        context = BridgeContext()
        context.prompt = item["prompt"]
        extra = {
            "force": True,
        }
        event = KernelEvent(KernelEvent.INPUT_SYSTEM, {
            'context': context,
            'extra': extra,
        })
        self.window.dispatch(event)

    def schedule_tasks(self):
        """Schedule tasks based on crontab"""
        crontab = self.get_option_value("crontab")
        # remove unused or inactive items
        for timer in self.timers:
            if not timer["item"]["enabled"] or timer["item"] not in crontab:
                self.timers.remove(timer)

        for item in crontab:
            try:
                if not item["enabled"]:
                    continue
                cron = item["crontab"]
                base_time = datetime.now()
                iter = croniter(cron, base_time)
                is_timer = False
                timer = None
                for timer in self.timers:
                    if item == timer["item"]:
                        is_timer = True
                        break

                # add to timers if not exists
                if not is_timer:
                    timer = {
                        "item": item,
                        "next_time": None,
                    }
                    self.timers.append(timer)

                next_time = timer["next_time"]

                # if first run
                if next_time is None:
                    next_time = iter.get_next(datetime)
                    timer["next_time"] = next_time
                    continue

                if datetime.now() >= next_time:
                    next_time = iter.get_next(datetime)
                    timer["next_time"] = next_time
                    self.job(item)

                timer["next_time"] = iter.get_next(datetime)
            except Exception as e:
                self.log("Error: {}".format(e))

        # show number of scheduled jobs
        num_jobs = len(self.timers)
        if num_jobs > 0:
            self.window.ui.plugin_addon['schedule'].setVisible(True)
            self.window.ui.plugin_addon['schedule'].setText(
                "+ Cron: {} job(s)".format(len(self.timers)),
            )
        else:
            self.window.ui.plugin_addon['schedule'].setVisible(False)
            self.window.ui.plugin_addon['schedule'].setText("")
